#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging


def setup_logging(level):
    """
    Set up the logging to use a decent format and the log level given as parameter.
    :param level: the log level used for the root logger
    """
    logging.basicConfig(format='%(asctime)s %(filename)s:%(lineno)04d %(levelname)s %(message)s')
    if level:
        getattr(logging, level.upper())
        numeric_level = getattr(logging, level.upper(), None)
        if not isinstance(numeric_level, int):
            raise ValueError('Invalid log level: %s' % level)
        logging.getLogger().setLevel(numeric_level)


def add_log_arguments(parser):
    """
    Adds command line arguments to control logging behaviour.

    :param parser: The argparse.ArgumentParser object to add to.
    """
    parser.add_argument('--log', action='store', dest='loglevel')



# -*- coding: UTF-8 -*-

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implements the ChaCha20 stream cipher and the Poly1350 authenticator. More information can be found on
https://tools.ietf.org/html/rfc7539. See HomeKit spec page 51.
"""
from math import ceil


def rotate_left(num: int, num_size: int, shift_bits: int) -> int:
    """
    Rotate a number num of num_size bits by shift_bits bits. See
    https://en.wikipedia.org/wiki/Bitwise_operation#Rotate_no_carry for more information.

    :param num: the number to rotate
    :param num_size: the size of the number in bits
    :param shift_bits: the number of bits the number is rotated by
    :return: the rotated number
    """
    mask = (2 ** shift_bits - 1) << (num_size - shift_bits)
    bits = (num & mask) >> (num_size - shift_bits)
    num = (num << shift_bits) & (2 ** num_size - 1)
    num |= bits
    return num


def chacha20_quarter_round(s: list, a: int, b: int, c: int, d: int):
    """
    Computes a chacha20 quarter round on the given state s as describe in RFC7539 Chapter 2.1. The state gets updated
    inplace.

    :param s: the state
    :param a: the first index into the state
    :param b: the second index into the state
    :param c: the third index into the state
    :param d: the fourth index into the state
    :return: None
    """
    assert len(s) == 16, 'state must have 16 ints'
    s[a] = (s[a] + s[b]) & (2 ** 32 - 1)
    s[d] = (s[d] ^ s[a])
    s[d] = rotate_left(s[d], 32, 16)

    s[c] = (s[c] + s[d]) & (2 ** 32 - 1)
    s[b] = (s[b] ^ s[c])
    s[b] = rotate_left(s[b], 32, 12)

    s[a] = (s[a] + s[b]) & (2 ** 32 - 1)
    s[d] = (s[d] ^ s[a])
    s[d] = rotate_left(s[d], 32, 8)

    s[c] = (s[c] + s[d]) & (2 ** 32 - 1)
    s[b] = (s[b] ^ s[c])
    s[b] = rotate_left(s[b], 32, 7)


def convert(num: int) -> int:
    return int.from_bytes(num.to_bytes(4, byteorder='little'), byteorder='big')


def extract_nth_word(num: bytes, w: int) -> int:
    m = len(num)
    bs = num[m - 4 * (w + 1):m - 4 * (w + 0)]
    return int.from_bytes(bs, byteorder='big')


def chacha20_create_initial_state(key: bytes, nonce: bytes, counter: int) -> list:
    """
    Creates the initial chacha20 state for the block function as described in RFC7539 chapter 2.3.

    :param key: the 256 bit key to use as bytes
    :param nonce: the 96 bit nonce as bytes
    :param counter: the 32bit block counter
    :return: the initial state as list of ints
    """
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32
    assert type(nonce) is bytes, 'nonce is no instance of bytes'
    assert len(nonce) == 3 * 32 / 8
    state = [0x61707865, 0x3320646e, 0x79622d32, 0x6b206574,
             0, 0, 0, 0,
             0, 0, 0, 0,
             0, 0, 0, 0]
    for i in range(0, 8):
        state[4 + i] = convert(extract_nth_word(key, 7 - i))
    state[12] = counter & (2 ** 32 - 1)
    state[13] = convert(extract_nth_word(nonce, 2))
    state[14] = convert(extract_nth_word(nonce, 1))
    state[15] = convert(extract_nth_word(nonce, 0))
    return state


def chacha20_inner_block(state: list):
    """
    The function to perform the inner quarter rounds as described in RFC7539 chapter 2.3.1. The state is updated
    inplace.

    :param state: the chacha20 state
    :return: None
    """
    assert len(state) == 16, 'state must have 16 ints'
    chacha20_quarter_round(state, 0, 4, 8, 12)
    chacha20_quarter_round(state, 1, 5, 9, 13)
    chacha20_quarter_round(state, 2, 6, 10, 14)
    chacha20_quarter_round(state, 3, 7, 11, 15)
    chacha20_quarter_round(state, 0, 5, 10, 15)
    chacha20_quarter_round(state, 1, 6, 11, 12)
    chacha20_quarter_round(state, 2, 7, 8, 13)
    chacha20_quarter_round(state, 3, 4, 9, 14)


def chacha20_block(key: bytes, nonce: bytes, counter: int) -> int:
    """
    The function to perform the computation for one chacha20 block as  described in RFC7539 chapter 2.3.1.

    :param key: the 256 bit key to use as bytes
    :param nonce: the 96 bit nonce as bytes
    :param counter: the 32bit block counter
    :return: the state serialized as int
    """
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32
    assert type(nonce) is bytes, 'nonce is no instance of bytes'
    assert len(nonce) == 3 * 32 / 8
    state = chacha20_create_initial_state(key, nonce, counter)
    w_state = state.copy()
    for i in range(0, 10):
        chacha20_inner_block(w_state)
    result = 0
    for i in range(0, 16):
        state[i] = (state[i] + w_state[i]) & (2 ** 32 - 1)
        result <<= 32
        result += convert(state[i])
    return result


def chacha20_encrypt(key: bytes, counter: int, nonce: int, plaintext: bytes) -> bytes:
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32
    assert type(nonce) is bytes, 'nonce is no instance of bytes'
    assert len(nonce) == 12
    encrypted = bytearray()
    for i in range(0, int(len(plaintext) / 64)):
        key_stream = chacha20_block(key, nonce, counter + i)
        key_stream = key_stream.to_bytes(64, byteorder='big')
        block = plaintext[i * 64:i * 64 + 64]
        encrypted += bytes([a ^ b for (a, b) in zip(block, key_stream)])

    if (len(plaintext) % 64) != 0:
        j = int(len(plaintext) / 64)
        key_stream = chacha20_block(key, nonce, counter + j)
        block = plaintext[(j * 64):len(plaintext)]
        key_stream = key_stream.to_bytes(64, byteorder='big')
        key_stream = key_stream[:len(block)]
        encrypted += bytes([a ^ b for (a, b) in zip(block, key_stream)])
    return encrypted


def clamp(r: int) -> int:
    tmp = r.to_bytes(length=16, byteorder='little')
    msk = 0x0ffffffc0ffffffc0ffffffc0fffffff.to_bytes(length=16, byteorder='big')
    res = bytearray()
    for i in range(0, len(tmp)):
        res.append(tmp[i] & msk[i])
    return int.from_bytes(res, byteorder='big')


def calc_r(k: int) -> int:
    assert type(k) is bytes, 'key is no instance of bytes'
    assert len(k) == 32
    tmp = k[0:16]
    return int.from_bytes(tmp, byteorder='big')


def calc_s(k: bytes) -> int:
    assert type(k) is bytes, 'key is no instance of bytes'
    assert len(k) == 32
    tmp = k[16:32]
    return int.from_bytes(tmp, byteorder='little')


def poly1305_mac(msg: bytes, key: bytes) -> bytes:
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32
    r = calc_r(key)
    r = clamp(r)
    s = calc_s(key)
    a = 0
    p = (1 << 130) - 5
    for i in range(0, ceil(len(msg) / 16)):
        block = bytearray(msg[i * 16:i * 16 + 16])
        block.append(0x01)
        n = int.from_bytes(block, byteorder='little')
        a += n
        a = (r * a)
        a %= p
    a += s
    a &= (2 ** (16 * 8) - 1)
    return a.to_bytes(length=16, byteorder='little')


def poly1305_key_gen(key: bytes, nonce: bytes) -> bytes:
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32
    assert type(nonce) is bytes, 'nonce is no instance of bytes'
    assert len(nonce) == 12
    counter = 0
    block = chacha20_block(key, nonce, counter)
    mask = (2 ** (32 * 8) - 1) << (32 * 8)
    block &= mask
    block >>= (32 * 8)
    block = block.to_bytes(length=32, byteorder='big')
    return block


def pad16(x: bytes) -> bytes:
    tmp = len(x) % 16
    if tmp == 0:
        return bytearray([])
    tmp = 16 - tmp
    return bytearray([0 for i in range(0, tmp)])


def chacha20_aead_verify_tag(aad: bytes, key: bytes, iv: bytes, constant: bytes, ciphertext: bytes):
    digest = ciphertext[-16:]
    ciphertext = ciphertext[:-16]

    nonce = constant + iv
    otk = poly1305_key_gen(key, nonce)
    mac_data = aad + pad16(aad)
    assert len(mac_data) % 16 == 0
    mac_data += ciphertext + pad16(ciphertext)
    assert len(mac_data) % 16 == 0
    mac_data += len(aad).to_bytes(length=8, byteorder='little')
    mac_data += len(ciphertext).to_bytes(length=8, byteorder='little')
    tag = poly1305_mac(mac_data, otk)

    return digest == tag


def chacha20_aead_encrypt(aad: bytes, key: bytes, iv: bytes, constant: bytes, plaintext: bytes):
    """
    The encrypt method for chacha20 aead as required by the Apple specification. The 96-bit nonce from RFC7539 is
    formed from the constant and the initialisation vector.

    :param aad: arbitrary length additional authenticated data
    :param key: 256-bit (32-byte) key of type bytes
    :param iv: the initialisation vector
    :param constant: constant
    :param plaintext: arbitrary length plaintext of type bytes or bytearray
    :return: the cipher text and tag
    """
    assert type(plaintext) in [bytes, bytearray], 'plaintext is no instance of bytes: %s' % str(type(plaintext))
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32

    nonce = constant + iv
    otk = poly1305_key_gen(key, nonce)
    ciphertext = chacha20_encrypt(key, 1, nonce, plaintext)
    assert len(plaintext) == len(ciphertext)
    mac_data = aad + pad16(aad)
    assert len(mac_data) % 16 == 0
    mac_data += ciphertext + pad16(ciphertext)
    assert len(mac_data) % 16 == 0
    mac_data += len(aad).to_bytes(length=8, byteorder='little')
    mac_data += len(ciphertext).to_bytes(length=8, byteorder='little')
    tag = poly1305_mac(mac_data, otk)
    return ciphertext, tag


def chacha20_aead_decrypt(aad: bytes, key: bytes, iv: bytes, constant: bytes, ciphertext: bytes):
    """
    The decrypt method for chacha20 aead as required by the Apple specification. The 96-bit nonce from RFC7539 is
    formed from the constant and the initialisation vector.

    :param aad: arbitrary length additional authenticated data
    :param key: 256-bit (32-byte) key of type bytes
    :param iv: the initialisation vector
    :param constant: constant
    :param ciphertext: arbitrary length plaintext of type bytes or bytearray
    :return: False if the tag could not be verified or the plaintext as bytes
    """
    assert type(ciphertext) in [bytes, bytearray], 'ciphertext is no instance of bytes: %s' % str(type(ciphertext))
    assert type(key) is bytes, 'key is no instance of bytes'
    assert len(key) == 32

    # break up on difference
    if not chacha20_aead_verify_tag(aad, key, iv, constant, ciphertext):
        return False

    # decrypt and return
    ciphertext = ciphertext[:-16]
    nonce = constant + iv
    plaintext = chacha20_encrypt(key, 1, nonce, ciphertext)
    assert len(plaintext) == len(ciphertext)
    return plaintext



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import binascii
import hashlib
import io
import json
from json.decoder import JSONDecodeError
import select

from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import ThreadingMixIn

import hkdf
from zeroconf import Zeroconf, ServiceInfo
import socket
import sys
import logging
import ed25519

from cryptography.hazmat.primitives.asymmetric import x25519
from cryptography.hazmat.primitives import serialization

from homekit.crypto.chacha20poly1305 import chacha20_aead_decrypt, chacha20_aead_encrypt
from homekit.crypto.srp import SrpServer

from homekit.exceptions import ConfigurationError, ConfigLoadingError, ConfigSavingError, FormatError, \
    CharacteristicPermissionError
from homekit.http_impl import HttpStatusCodes
from homekit.model import Accessories, Categories
from homekit.model.characteristics import CharacteristicsTypes
from homekit.protocol import TLV
from homekit.protocol.statuscodes import HapStatusCodes


class AccessoryServer(ThreadingMixIn, HTTPServer):
    """
    This server makes accessories accessible via the HomeKit protocol.
    """

    def __init__(self, config_file, logger=sys.stderr):
        """
        Create a new server that acts like a homekit accessory. The config file is loaded and checked.

        :param config_file: the file that contains the configuration data. Must be a string representing an absolute
        path to the file
        :param logger: this can be None to disable logging, sys.stderr to use the default behaviour of the python
        implementation or an instance of logging.Logger to use this.
        :raises HomeKitConfigurationException: if the config file is malformed. Reason will be in the message.
        """
        if logger is None or logger == sys.stderr or isinstance(logger, logging.Logger):
            self.logger = logger
        else:
            raise ConfigurationError('Invalid logger given.')

        self.data = AccessoryServerData(config_file)
        self.data.increase_configuration_number()
        self.sessions = {}
        self.zeroconf = Zeroconf()
        self.mdns_type = '_hap._tcp.local.'
        self.mdns_name = self.data.name + '._hap._tcp.local.'
        self.identify_callback = None
        self.zeroconf_info = None

        self.accessories = Accessories()

        HTTPServer.__init__(self, (self.data.ip, self.data.port), AccessoryRequestHandler)

    def add_accessory(self, accessory):
        self.accessories.add_accessory(accessory)

    def set_identify_callback(self, func):
        """
        Sets the callback function for this accessory server. This will NOT be applied to all accessories registered
        with. This function will be called on unpaired calls to identify.

        :param func: a function without any parameters and without return type.
        """
        self.identify_callback = func

    def publish_device(self):
        desc = {'md': str(self.data.name),  # model name of accessory
                # category identifier (page 254, 2 means bridge), must be a String
                'ci': str(Categories[self.data.category]),
                'pv': '1.0',  # protocol version
                'c#': str(self.data.configuration_number),
                # configuration (consecutive number, 1 or greater, must be changed on every configuration change)
                'id': self.data.accessory_pairing_id_bytes,  # id MUST look like Mac Address
                'ff': '0',  # feature flags (Table 5-8, page 69)
                's#': '1',  # must be 1
                'sf': '1'  # status flag, lowest bit encodes pairing status, 1 means unpaired
                }
        if self.data.is_paired:
            desc['sf'] = '0'

        self.zeroconf_info = ServiceInfo(self.mdns_type, self.mdns_name, address=socket.inet_aton(self.data.ip),
                                         port=self.data.port, properties=desc)
        self.zeroconf.unregister_service(self.zeroconf_info)
        self.zeroconf.register_service(self.zeroconf_info, allow_name_change=True)

    def unpublish_device(self):
        if self.zeroconf_info:
            self.zeroconf.unregister_service(self.zeroconf_info)

    def shutdown(self):
        # tell all handlers to close the connection
        for session in self.sessions:
            self.sessions[session]['handler'].close_connection = True
        self.socket.close()
        HTTPServer.shutdown(self)


class AccessoryServerData:
    """
    This class is used to take care of the servers persistence to be able to manage restarts,
    """

    def __init__(self, data_file):
        self.data_file = data_file
        try:
            with open(data_file, 'r') as input_file:
                self.data = json.load(input_file)
        except PermissionError:
            raise ConfigLoadingError('Could not open "{f}" due to missing permissions'.format(f=input_file))
        except JSONDecodeError:
            raise ConfigLoadingError('Cannot parse "{f}" as JSON file'.format(f=input_file))
        except FileNotFoundError:
            raise ConfigLoadingError('Could not open "{f}" because it does not exist'.format(f=input_file))

        self.check()

        # set some default values
        if 'peers' not in self.data:
            self.data['peers'] = {}
        if 'unsuccessful_tries' not in self.data:
            self.data['unsuccessful_tries'] = 0

    def _save_data(self):
        try:
            with open(self.data_file, 'w') as output_file:
                json.dump(self.data, output_file, indent=2, sort_keys=True)
        except PermissionError:
            raise ConfigSavingError('Could not write "{f}" due to missing permissions'.format(f=self.data_file))
        except FileNotFoundError:
            raise ConfigSavingError(
                'Could not write "{f}" because it (or the folder) does not exist'.format(f=self.data_file))

    @property
    def ip(self) -> str:
        return self.data['host_ip']

    @property
    def port(self) -> int:
        return self.data['host_port']

    @property
    def setup_code(self) -> str:
        return self.data['accessory_pin']

    @property
    def accessory_pairing_id_bytes(self) -> bytes:
        return self.data['accessory_pairing_id'].encode()

    @property
    def unsuccessful_tries(self) -> int:
        return self.data['unsuccessful_tries']

    def register_unsuccessful_try(self):
        self.data['unsuccessful_tries'] += 1
        self._save_data()

    @property
    def is_paired(self) -> bool:
        return len(self.data['peers']) > 0

    @property
    def name(self) -> str:
        return self.data['name']

    @property
    def category(self) -> str:
        try:
            category = self.data['category']
        except KeyError:
            raise ConfigurationError('category missing in "{f}"'.format(f=self.data_file))
        if category not in Categories:
            raise ConfigurationError('invalid category "{c}" in "{f}"'.format(c=category, f=self.data_file))
        return category

    def remove_peer(self, pairing_id: bytes):
        del self.data['peers'][pairing_id.decode()]
        self._save_data()

    def set_peer_permissions(self, pairing_id: bytes, admin: bool):
        peer = self.data['peers'][pairing_id.decode()]
        peer['admin'] = admin
        self._save_data()

    def add_peer(self, pairing_id: bytes, ltpk: bytes, admin: bool):
        self.data['peers'][pairing_id.decode()] = {'key': binascii.hexlify(ltpk).decode(), 'admin': admin}
        self._save_data()

    def get_peer_key(self, pairing_id: bytes) -> bytes:
        if pairing_id.decode() in self.data['peers']:
            return bytes.fromhex(self.data['peers'][pairing_id.decode()]['key'])
        else:
            return None

    def is_peer_admin(self, pairing_id: bytes) -> bool:
        return self.data['peers'][pairing_id.decode()]['admin']

    @property
    def peers(self):
        return self.data['peers'].keys()

    @property
    def accessory_ltsk(self) -> bytes:
        if 'accessory_ltsk' in self.data:
            return bytes.fromhex(self.data['accessory_ltsk'])
        else:
            return None

    @property
    def accessory_ltpk(self) -> bytes:
        if 'accessory_ltpk' in self.data:
            return bytes.fromhex(self.data['accessory_ltpk'])
        else:
            return None

    def set_accessory_keys(self, accessory_ltpk: bytes, accessory_ltsk: bytes):
        self.data['accessory_ltpk'] = binascii.hexlify(accessory_ltpk).decode()
        self.data['accessory_ltsk'] = binascii.hexlify(accessory_ltsk).decode()[:64]
        self._save_data()

    @property
    def configuration_number(self) -> int:
        return self.data['c#']

    def increase_configuration_number(self):
        self.data['c#'] += 1
        self._save_data()

    def check(self, paired=False):
        """
        Checks a accessory config file for completeness.
        :param paired: if True, check for keys that must exist after successful pairing as well.
        :return: None, but a HomeKitConfigurationException is raised if a key is missing
        """
        required_fields = ['name', 'host_ip', 'host_port', 'accessory_pairing_id', 'accessory_pin', 'c#', 'category']
        if paired:
            required_fields.extend(['accessory_ltpk', 'accessory_ltsk', 'peers', 'unsuccessful_tries'])
        for f in required_fields:
            if f not in self.data:
                raise ConfigurationError(
                    '"{r}" is missing in the config file "{f}"!'.format(r=f, f=self.data_file))


class AccessoryRequestHandler(BaseHTTPRequestHandler):
    VALID_METHODS = ['GET', 'HEAD', 'POST', 'PUT', 'DELETE', 'CONNECT', 'OPTIONS', 'TRACE']
    DEBUG_PUT_CHARACTERISTICS = False
    DEBUG_CRYPT = False
    DEBUG_PAIR_VERIFY = False
    DEBUG_GET_CHARACTERISTICS = False
    timeout = 300

    def __init__(self, request, client_address, server):
        # keep pycharm from complaining about those not being define in __init__
        # self.session_id = '{ip}:{port}'.format(ip=client_address[0], port= client_address[1])
        self.session_id = '{ip}'.format(ip=client_address[0])
        if self.session_id not in server.sessions:
            server.sessions[self.session_id] = {'handler': self}
        self.rfile = None
        self.wfile = None
        self.body = None
        self.PATHMAPPING = {
            '/accessories': {
                'GET': self._get_accessories
            },
            '/characteristics': {
                'GET': self._get_characteristics,
                'PUT': self._put_characteristics
            },
            '/identify': {
                'POST': self._post_identify
            },
            '/pair-setup': {
                'POST': self._post_pair_setup
            },
            '/pair-verify': {
                'POST': self._post_pair_verify
            },
            '/pairings': {
                'POST': self._post_pairings
            }
        }
        self.protocol_version = 'HTTP/1.1'
        self.close_connection = False

        self.timeout_counter = 0

        # get the identify callback function from calling server
        self.identify_callback = server.identify_callback

        # init super class
        BaseHTTPRequestHandler.__init__(self, request, client_address, server)

    def handle_one_request(self):
        """
        This is used to determine whether the request is encrypted or not. This is done by looking at the first bytes of
        the request. To be valid unencrypted HTTP call, it must be one of the methods defined in RFC7231 Section 4
        "Request Methods".
        """
        try:
            # make connection non blocking so the select can work
            self.connection.setblocking(0)
            ready = select.select([self.connection], [], [], 1)

            # no data was to be received, so we count up to track how many seconds in total this happened
            if not ready[0]:
                self.timeout_counter += 1

                # if this is above our configured timeout the connection gets closed
                if self.timeout_counter >= self.timeout:
                    self.close_connection = True
                return

            raw_peeked_data = self.rfile.peek(10)
            if len(raw_peeked_data) == 0:
                # since select says ready but no data is there, close the connection to prevent hidden busy waiting to
                # rise load
                self.close_connection = True
                return

            # data was received so reset the timeout handler
            self.timeout_counter = 0

            # RFC7230 Section 3 tells us, that US-ASCII is fine
            peeked_data = raw_peeked_data[:10]
            peeked_data = peeked_data.decode(encoding='ASCII')
            # self.log_message('deciding over: >%s<', peeked_data)
            # If the request line starts with a known HTTP verb, then use handle_one_request from super class
            if ' ' in peeked_data:
                method = peeked_data.split(' ', 1)[0]
                if method in self.VALID_METHODS:
                    self.server.sessions[self.session_id]['enrypted_connection'] = False
                    BaseHTTPRequestHandler.handle_one_request(self)
                    return
        except (socket.timeout, OSError) as e:
            # a read or a write timed out.  Discard this connection
            self.log_error(' %r', e)
            self.close_connection = True
            return
        except UnicodeDecodeError as e:
            # this just means it might be encrypted...
            self.log_debug('Unicode exception %s' % e)
            pass

        # the first 2 bytes are the length of the encrypted data to follow
        len_bytes = self.rfile.read(2)
        data_len = int.from_bytes(len_bytes, byteorder='little')

        # the auth tag is not counted, so add its length
        data = self.rfile.read(data_len + 16)
        if AccessoryRequestHandler.DEBUG_CRYPT:
            self.log_message('data >%i< >%s<', len(data), binascii.hexlify(data))

        # get the crypto key from the session
        c2a_key = self.server.sessions[self.session_id]['controller_to_accessory_key']

        # verify & decrypt the read data
        cnt_bytes = self.server.sessions[self.session_id]['controller_to_accessory_count'].to_bytes(8,
                                                                                                    byteorder='little')
        decrypted = chacha20_aead_decrypt(len_bytes, c2a_key, cnt_bytes, bytes([0, 0, 0, 0]),
                                          data)
        if decrypted is False:
            # crypto error, log it and request close of connection
            self.log_error('SEVERE: Could not decrypt %s', binascii.hexlify(data))
            self.close_connection = True
            return

        if AccessoryRequestHandler.DEBUG_CRYPT:
            self.log_message('crypted request >%s<', decrypted)

        self.server.sessions[self.session_id]['controller_to_accessory_count'] += 1

        # replace the original rfile with a fake with the decrypted stuff
        old_rfile = self.rfile
        self.rfile = io.BytesIO(decrypted)

        # replace writefile to pass on encrypted data
        old_wfile = self.wfile
        self.wfile = io.BytesIO()

        # call known function
        self.server.sessions[self.session_id]['enrypted_connection'] = True
        BaseHTTPRequestHandler.handle_one_request(self)

        # read the plaintext and send it out encrypted
        self.wfile.seek(0)
        in_data = self.wfile.read(65537)

        if AccessoryRequestHandler.DEBUG_CRYPT:
            self.log_message('response >%s<', in_data)
            self.log_message('len(response) %s', len(in_data))

        block_size = 1024
        out_data = bytearray()
        while len(in_data) > 0:
            block = in_data[:block_size]
            if AccessoryRequestHandler.DEBUG_CRYPT:
                self.log_message('==> BLOCK: len %s', len(block))
            in_data = in_data[block_size:]

            len_bytes = len(block).to_bytes(2, byteorder='little')
            a2c_key = self.server.sessions[self.session_id]['accessory_to_controller_key']
            cnt_bytes = self.server.sessions[self.session_id]['accessory_to_controller_count'].\
                to_bytes(8, byteorder='little')
            ciper_and_mac = chacha20_aead_encrypt(len_bytes, a2c_key, cnt_bytes, bytes([0, 0, 0, 0]), block)
            self.server.sessions[self.session_id]['accessory_to_controller_count'] += 1
            out_data += len_bytes + ciper_and_mac[0] + ciper_and_mac[1]

        # change back to originals to handle multiple calls
        self.rfile = old_rfile
        self.wfile = old_wfile

        # send data to original requester
        self.wfile.write(out_data)
        self.wfile.flush()

    def _get_characteristics(self):
        """
        As described on page 84
        :return:
        """
        if AccessoryRequestHandler.DEBUG_GET_CHARACTERISTICS:
            self.log_message('GET /characteristics')

        # analyse
        params = {}
        if '?' in self.path:
            params = {t.split('=')[0]: t.split('=')[1] for t in self.path.split('?')[1].split('&')}

        # handle id param
        ids = []
        if 'id' in params:
            ids = params['id'].split(',')

        # handle meta param
        meta = False
        if 'meta' in params:
            meta = params['meta'] == '1'

        # handle perms param
        perms = False
        if 'perms' in params:
            perms = params['perms'] == '1'

        # handle type param
        include_type = False
        if 'type' in params:
            include_type = params['type'] == '1'

        # handle ev param
        ev = False
        if 'ev' in params:
            ev = params['ev'] == '1'

        if AccessoryRequestHandler.DEBUG_GET_CHARACTERISTICS:
            self.log_message('query parameters: ids: %s, meta: %s, perms: %s, type: %s, ev: %s', ids, meta, perms,
                             include_type, ev)

        result = {
            'characteristics': []
        }

        errors = 0
        for id_pair in ids:
            id_pair = id_pair.split('.')
            aid = int(id_pair[0])
            cid = int(id_pair[1])
            found = False
            for accessory in self.server.accessories.accessories:
                if accessory.aid != aid:
                    continue
                for service in accessory.services:
                    for characteristic in service.characteristics:
                        if characteristic.iid != cid:
                            continue
                        found = True
                        # try to read the characteristic and report possible exceptions as error
                        try:
                            value = characteristic.get_value()
                            result['characteristics'].append({'aid': aid, 'iid': cid, 'value': value})
                        except FormatError:
                            result['characteristics'].append(
                                {'aid': aid, 'iid': cid, 'status': HapStatusCodes.INVALID_VALUE})
                            errors += 1
                        except CharacteristicPermissionError:
                            result['characteristics'].append(
                                {'aid': aid, 'iid': cid, 'status': HapStatusCodes.CANT_READ_WRITE_ONLY})
                            errors += 1
                        except Exception as e:
                            self.log_error('Exception while getting value for %s.%s: %s', aid, cid, str(e))
                            result['characteristics'].append(
                                {'aid': aid, 'iid': cid, 'status': HapStatusCodes.OUT_OF_RESOURCES})
                            errors += 1
                        if ev:
                            # TODO handling of events is missing
                            result['characteristics'][-1]['ev'] = False
                        if include_type:
                            result['characteristics'][-1]['type'] = \
                                CharacteristicsTypes.get_short_uuid(characteristic.type)
                        if perms:
                            result['characteristics'][-1]['perms'] = characteristic.perms
                        if meta:
                            meta_data = characteristic.get_meta()
                            for key in meta_data:
                                result['characteristics'][-1][key] = meta_data[key]
                # report missing resources
                if not found:
                    result['characteristics'].append(
                        {'aid': aid, 'iid': cid, 'status': HapStatusCodes.RESOURCE_NOT_EXIST})
                    errors += 1

        if AccessoryRequestHandler.DEBUG_GET_CHARACTERISTICS:
            self.log_message('chars: %s', json.dumps(result))

        # set the proper status code depending on the count of characteristics and error
        if len(result['characteristics']) == errors:
            self.send_response(HttpStatusCodes.BAD_REQUEST)
        elif len(result['characteristics']) > errors > 0:
            self.send_response(HttpStatusCodes.MULTI_STATUS)
        else:
            self.send_response(HttpStatusCodes.OK)

        result_bytes = json.dumps(result).encode()
        self.send_header('Content-Type', 'application/hap+json')
        self.send_header('Content-Length', len(result_bytes))
        self.end_headers()
        self.wfile.write(result_bytes)

    def _put_characteristics(self):
        """
        Defined page 80 ff
        :return:
        """
        if AccessoryRequestHandler.DEBUG_PUT_CHARACTERISTICS:
            self.log_message('PUT /characteristics')
            self.log_message('body: %s', self.body)

        data = json.loads(self.body.decode())
        characteristics_to_set = data['characteristics']
        result = {
            'characteristics': []
        }
        errors = 0
        for characteristic_to_set in characteristics_to_set:
            aid = characteristic_to_set['aid']
            cid = characteristic_to_set['iid']
            found = False
            for accessory in self.server.accessories.accessories:
                if accessory.aid != aid:
                    continue
                for service in accessory.services:
                    for characteristic in service.characteristics:
                        if characteristic.iid != cid:
                            continue
                        found = True
                        if 'ev' in characteristic_to_set:
                            if AccessoryRequestHandler.DEBUG_PUT_CHARACTERISTICS:
                                self.log_message('set ev >%s< >%s< >%s<', aid, cid, characteristic_to_set['ev'])
                            if 'ev' in characteristic.perms:
                                characteristic.set_events(characteristic_to_set['ev'])
                                result['characteristics'].append({'aid': aid, 'iid': cid, 'status': 0})
                            else:
                                result['characteristics'].append(
                                    {'aid': aid, 'iid': cid, 'status': HapStatusCodes.NOTIFICATION_NOT_SUPPORTED})

                        if 'value' in characteristic_to_set:
                            if AccessoryRequestHandler.DEBUG_PUT_CHARACTERISTICS:
                                self.log_message('set value >%s< >%s< >%s<', aid, cid, characteristic_to_set['value'])
                            try:
                                characteristic.set_value(characteristic_to_set['value'])
                                result['characteristics'].append({'aid': aid, 'iid': cid, 'status': 0})
                            except FormatError:
                                result['characteristics'].append(
                                    {'aid': aid, 'iid': cid, 'status': HapStatusCodes.INVALID_VALUE})
                                errors += 1
                            except Exception as e:
                                self.log_error('Exception while setting value for %s.%s: %s', aid, cid, str(e))
                                result['characteristics'].append(
                                    {'aid': aid, 'iid': cid, 'status': HapStatusCodes.OUT_OF_RESOURCES})
                                errors += 1
            # report missing resources
            if not found:
                result['characteristics'].append(
                    {'aid': aid, 'iid': cid, 'status': HapStatusCodes.RESOURCE_NOT_EXIST})
                errors += 1

        if len(result['characteristics']) == errors:
            self.send_response(HttpStatusCodes.BAD_REQUEST)
        elif len(result['characteristics']) > errors > 0:
            self.send_response(HttpStatusCodes.MULTI_STATUS)
        else:
            self.send_response(HttpStatusCodes.NO_CONTENT)
            self.end_headers()
        if errors > 0:
            result_bytes = json.dumps(result).encode()
            self.send_header('Content-Type', 'application/hap+json')
            self.send_header('Content-Length', len(result_bytes))
            self.end_headers()
            self.wfile.write(result_bytes)

    def _post_identify(self):
        if self.server.data.is_paired:
            result_bytes = json.dumps({'status': HapStatusCodes.INSUFFICIENT_PRIVILEGES}).encode()
            self.send_response(HttpStatusCodes.BAD_REQUEST)
            self.send_header('Content-Type', 'application/hap+json')
            self.send_header('Content-Length', len(result_bytes))
            self.end_headers()
            self.wfile.write(result_bytes)
        else:
            # perform identify action
            if self.identify_callback:
                self.identify_callback()
            # send status code
            self.send_response(HttpStatusCodes.NO_CONTENT)
            self.end_headers()

    def _get_accessories(self):

        result_bytes = self.server.accessories.to_accessory_and_service_list().encode()
        self.send_response(HttpStatusCodes.OK)
        self.send_header('Content-Type', 'application/hap+json')
        self.send_header('Content-Length', len(result_bytes))
        self.end_headers()
        self.wfile.write(result_bytes)

    def _post_pair_verify(self):
        d_req = TLV.decode_bytes(self.body)

        # Order is not consistent, so force things in to order specified in spec
        d_req = TLV.reorder(d_req, [
            TLV.kTLVType_State,
            TLV.kTLVType_PublicKey,
            TLV.kTLVType_EncryptedData,
        ])

        d_res = []

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M1:
            # step #2 Accessory -> iOS Device Verify Start Response
            if AccessoryRequestHandler.DEBUG_PAIR_VERIFY:
                self.log_message('Step #2 /pair-verify')

            # 1) generate new curve25519 key pair
            accessory_session_key = x25519.X25519PrivateKey.generate()
            accessory_spk = accessory_session_key.public_key().public_bytes(
                encoding=serialization.Encoding.Raw,
                format=serialization.PublicFormat.Raw
            )
            self.server.sessions[self.session_id]['accessory_pub_key'] = accessory_spk

            # 2) generate shared secret
            ios_device_curve25519_pub_key_bytes = bytes(d_req[1][1])
            self.server.sessions[self.session_id]['ios_device_pub_key'] = ios_device_curve25519_pub_key_bytes
            ios_device_curve25519_pub_key = x25519.X25519PublicKey.from_public_bytes(
                ios_device_curve25519_pub_key_bytes)

            shared_secret = accessory_session_key.exchange(ios_device_curve25519_pub_key)
            self.server.sessions[self.session_id]['shared_secret'] = shared_secret

            # 3) generate accessory info
            accessory_info = accessory_spk + self.server.data.accessory_pairing_id_bytes + \
                ios_device_curve25519_pub_key_bytes

            # 4) sign accessory info for accessory signature
            # accessory_ltsk = ed25519.SigningKey(self.server.data.accessory_ltsk + self.server.data.accessory_ltpk)
            accessory_ltsk = ed25519.SigningKey(self.server.data.accessory_ltsk)
            accessory_signature = accessory_ltsk.sign(accessory_info)

            # 5) sub tlv
            sub_tlv = [
                (TLV.kTLVType_Identifier, self.server.data.accessory_pairing_id_bytes),
                (TLV.kTLVType_Signature, accessory_signature)
            ]
            sub_tlv_b = TLV.encode_list(sub_tlv)

            # 6) derive session key
            hkdf_inst = hkdf.Hkdf('Pair-Verify-Encrypt-Salt'.encode(), shared_secret, hash=hashlib.sha512)
            session_key = hkdf_inst.expand('Pair-Verify-Encrypt-Info'.encode(), 32)
            self.server.sessions[self.session_id]['session_key'] = session_key

            # 7) encrypt sub tlv
            encrypted_data_with_auth_tag = chacha20_aead_encrypt(bytes(),
                                                                 session_key,
                                                                 'PV-Msg02'.encode(),
                                                                 bytes([0, 0, 0, 0]),
                                                                 sub_tlv_b)
            tmp = bytearray(encrypted_data_with_auth_tag[0])
            tmp += encrypted_data_with_auth_tag[1]

            # 8) construct result tlv
            d_res.append((TLV.kTLVType_State, TLV.M2,))
            d_res.append((TLV.kTLVType_PublicKey, accessory_spk))
            d_res.append((TLV.kTLVType_EncryptedData, tmp))

            self._send_response_tlv(d_res)
            if AccessoryRequestHandler.DEBUG_PAIR_VERIFY:
                self.log_message('after step #2\n%s', TLV.to_string(d_res))
            return

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M3:
            # step #4 Accessory -> iOS Device Verify Finish Response
            if AccessoryRequestHandler.DEBUG_PAIR_VERIFY:
                self.log_message('Step #4 /pair-verify')

            session_key = self.server.sessions[self.session_id]['session_key']

            # 1) verify ios' authtag
            # 2) decrypt
            encrypted = d_req[1][1]
            decrypted = chacha20_aead_decrypt(bytes(), session_key, 'PV-Msg03'.encode(), bytes([0, 0, 0, 0]),
                                              encrypted)
            if decrypted is False:
                self.send_error_reply(TLV.M4, TLV.kTLVError_Authentication)
                self.log_error('error in step #4: authtag %s %s', d_res, self.server.sessions)
                return
            d1 = TLV.decode_bytes(decrypted)
            assert d1[0][0] == TLV.kTLVType_Identifier
            assert d1[1][0] == TLV.kTLVType_Signature

            # 3) get ios_device_ltpk
            ios_device_pairing_id = d1[0][1]
            self.server.sessions[self.session_id]['ios_device_pairing_id'] = ios_device_pairing_id
            ios_device_ltpk_bytes = self.server.data.get_peer_key(ios_device_pairing_id)
            if ios_device_ltpk_bytes is None:
                self.send_error_reply(TLV.M4, TLV.kTLVError_Authentication)
                self.log_error('error in step #4: not paired %s %s', d_res, self.server.sessions)
                return
            ios_device_lptk = ed25519.VerifyingKey(ios_device_ltpk_bytes)

            # 4) verify ios_device_info
            ios_device_sig = d1[1][1]
            ios_device_curve25519_pub_key_bytes = self.server.sessions[self.session_id]['ios_device_pub_key']
            accessory_spk = self.server.sessions[self.session_id]['accessory_pub_key']
            ios_device_info = ios_device_curve25519_pub_key_bytes + ios_device_pairing_id + accessory_spk
            try:
                ios_device_lptk.verify(bytes(ios_device_sig), bytes(ios_device_info))
            except ed25519.BadSignatureError:
                self.send_error_reply(TLV.M4, TLV.kTLVError_Authentication)
                self.log_error('error in step #4: signature %s %s', d_res, self.server.sessions)
                return

            #
            shared_secret = self.server.sessions[self.session_id]['shared_secret']
            hkdf_inst = hkdf.Hkdf('Control-Salt'.encode(), shared_secret, hash=hashlib.sha512)
            controller_to_accessory_key = hkdf_inst.expand('Control-Write-Encryption-Key'.encode(), 32)
            self.server.sessions[self.session_id]['controller_to_accessory_key'] = controller_to_accessory_key
            self.server.sessions[self.session_id]['controller_to_accessory_count'] = 0

            hkdf_inst = hkdf.Hkdf('Control-Salt'.encode(), shared_secret, hash=hashlib.sha512)
            accessory_to_controller_key = hkdf_inst.expand('Control-Read-Encryption-Key'.encode(), 32)
            self.server.sessions[self.session_id]['accessory_to_controller_key'] = accessory_to_controller_key
            self.server.sessions[self.session_id]['accessory_to_controller_count'] = 0

            d_res.append((TLV.kTLVType_State, TLV.M4,))

            self._send_response_tlv(d_res)
            if AccessoryRequestHandler.DEBUG_PAIR_VERIFY:
                self.log_message('after step #4\n%s', TLV.to_string(d_res))
            return

        self.send_error(HttpStatusCodes.METHOD_NOT_ALLOWED)

    def _post_pairings(self):
        d_req = TLV.decode_bytes(self.body)

        # Order is not consistent, so force things in to order specified in spec
        d_req = TLV.reorder(d_req, [
            TLV.kTLVType_State,
            TLV.kTLVType_Method,

            # AddPairing/RemovePairing, M1
            TLV.kTLVType_Identifier,

            # AddPairing, M1
            TLV.kTLVType_PublicKey,
            TLV.kTLVType_Permissions,
        ])

        self.log_message('POST /pairings request body:\n%s', TLV.to_string(d_req))

        session = self.server.sessions[self.session_id]
        server_data = self.server.data

        d_res = []

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M1 \
                and d_req[1][0] == TLV.kTLVType_Method and d_req[1][1] == TLV.AddPairing:
            self.log_message('Step #2 /pairings add pairing')
            d_res.append((TLV.kTLVType_State, TLV.M2,))

            # see page 51
            # 1)

            # 2) verify admin bit is set
            ios_device_pairing_id = session['ios_device_pairing_id']
            if not server_data.is_peer_admin(ios_device_pairing_id):
                self.send_error_reply(TLV.M2, TLV.kTLVError_Authentication)
                self.log_error('error in step #2: admin bit')
                return

            additional_controller_pairing_identifier = d_req[2][1]
            additional_controller_LTPK = d_req[3][1]
            additional_controller_permissions = d_req[4][1]
            is_admin = additional_controller_permissions == b'\x01'

            # 3) pairing exists?
            registered_controller_LTPK = server_data.get_peer_key(additional_controller_pairing_identifier)

            if registered_controller_LTPK is not None:
                self.log_message('controller is already registered!')
                if registered_controller_LTPK != additional_controller_LTPK:
                    self.log_message('with different key')
                    # 3.a)
                    self.send_error_reply(TLV.M2, TLV.kTLVError_Unknown)
                    return

                self.log_message('with different permissions')
                # 3.b) update permission
                server_data.set_peer_permissions(additional_controller_pairing_identifier, is_admin)
            else:
                self.log_message('add pairing')

                # 4) no pairing exists
                # 4.a) no limit applied to number of pairings
                # 4.b) add pairing (could raise kTLVError_Unknown)
                server_data.add_peer(additional_controller_pairing_identifier, additional_controller_LTPK, is_admin)

            self._send_response_tlv(d_res)
            self.log_message('after step #2\n%s', TLV.to_string(d_res))

            return

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M1 \
                and d_req[1][0] == TLV.kTLVType_Method and d_req[1][1] == TLV.RemovePairing:
            # step #2 Accessory -> iOS Device remove pairing response
            self.log_message('Step #2 /pairings remove pairings')

            # 1)

            # 2) verify set admin bit
            ios_device_pairing_id = session['ios_device_pairing_id']
            if not server_data.is_peer_admin(ios_device_pairing_id):
                self.send_error_reply(TLV.M2, TLV.kTLVError_Authentication)
                self.log_error('error in step #2: admin bit not set for controller %s', ios_device_pairing_id.decode())
                return

            # 3) remove pairing and republish device
            server_data.remove_peer(d_req[2][1])
            self.server.publish_device()

            d_res.append((TLV.kTLVType_State, TLV.M2,))
            self._send_response_tlv(d_res)
            self.log_message('after step #2\n%s', TLV.to_string(d_res))

            # 6) + 7) invalidate HAP session and close connections
            # TODO implement this in more details
            # for session_id in self.server.sessions:
            #    session = self.server.sessions[session_id]
            #    if session['ios_device_pairing_id'] == d_req[TLV.kTLVType_Identifier]:
            #        session['handler'].close_connection = True
            #
            # if self.server.sessions[self.session_id]['ios_device_pairing_id'] == d_req[TLV.kTLVType_Identifier]:
            #    self.close_connection = True
            return

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M1 \
                and d_req[1][0] == TLV.kTLVType_Method and d_req[1][1] == TLV.ListPairings:
            # step #2 Accessory -> iOS Device list pairing response
            self.log_message('Step #2 /pairings list pairings')

            # 1) Validate against session

            # 2) verify set admin bit
            ios_device_pairing_id = session['ios_device_pairing_id']
            if not server_data.is_peer_admin(ios_device_pairing_id):
                self.send_error_reply(TLV.M2, TLV.kTLVError_Authentication)
                self.log_error('error in step #2: admin bit')
                return

            # 3) construct response TLV
            tmp = [(TLV.kTLVType_State, TLV.M2)]
            for index, pairing_id in enumerate(server_data.peers):
                tmp.append((TLV.kTLVType_Identifier, pairing_id.encode()))
                tmp.append((TLV.kTLVType_PublicKey, server_data.get_peer_key(pairing_id.encode())))
                user = TLV.kTLVType_Permission_RegularUser
                if server_data.is_peer_admin(pairing_id.encode()):
                    user = TLV.kTLVType_Permission_AdminUser
                tmp.append((TLV.kTLVType_Permissions, user))
                if index + 1 < len(server_data.peers):
                    tmp.append((TLV.kTLVType_Separator, bytes(0)))
            result_bytes = TLV.encode_list(tmp)

            # 4) send response
            self.send_response(HttpStatusCodes.OK)
            # Send headers
            self.send_header('Content-Length', len(result_bytes))
            self.send_header('Content-Type', 'application/pairing+tlv8')
            self.send_header('Connection', 'keep-alive')
            self.end_headers()

            self.wfile.write(result_bytes)
            return

        self.send_error(HttpStatusCodes.METHOD_NOT_ALLOWED)

    def send_error_reply(self, state, error):
        """
        Send an error reply encoded as TLV.
        :param state: The state as in TLV.M1, TLV.M2, ...
        :param error: The error code as in TLV.kTLVError_*
        :return: None
        """
        d_res = [
            (TLV.kTLVType_State, state),
            (TLV.kTLVType_Error, error)
        ]
        result_bytes = TLV.encode_list(d_res)

        self.send_response(HttpStatusCodes.METHOD_NOT_ALLOWED)
        # Send headers
        self.send_header('Content-Length', len(result_bytes))
        self.send_header('Content-Type', 'application/pairing+tlv8')
        self.end_headers()

        self.wfile.write(result_bytes)

    def _post_pair_setup(self):
        d_req = TLV.decode_bytes(self.body)

        # Order is not consistent, so force things in to order specified in spec
        d_req = TLV.reorder(d_req, [
            TLV.kTLVType_State,
            TLV.kTLVType_Method,
            # For M3
            TLV.kTLVType_PublicKey,
            TLV.kTLVType_Proof,
            # For M5
            TLV.kTLVType_EncryptedData,
        ])

        self.log_message('POST /pair-setup request body:\n%s', TLV.to_string(d_req))

        d_res = []

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M1:
            # step #2 Accessory -> iOS Device SRP Start Response
            self.log_message('Step #2 /pair-setup')

            # 1) Check if paired
            if self.server.data.is_paired:
                self.send_error_reply(TLV.M2, TLV.kTLVError_Unavailable)
                return

            # 2) Check if over 100 attempts
            if self.server.data.unsuccessful_tries > 100:
                self.log_error('to many failed attempts')
                self.send_error_reply(TLV.M2, TLV.kTLVError_MaxTries)
                return

            # 3) Check if already in pairing
            if False:
                self.send_error_reply(TLV.M2, TLV.kTLVError_Busy)
                return

            # 4) 5) 7) Create in SRP Session, set username and password
            server = SrpServer('Pair-Setup', self.server.data.setup_code)

            # 6) create salt
            salt = server.get_salt()

            # 8) show setup code to user
            sc = self.server.data.setup_code
            sc_str = 'Setup Code\n┌─' + '─' * len(sc) + '─┐\n│ ' + sc + ' │\n└─' + '─' * len(sc) + '─┘'
            self.log_message(sc_str)

            # 9) create public key
            public_key = server.get_public_key()

            # 10) create response tlv and send response
            d_res.append((TLV.kTLVType_State, TLV.M2,))
            d_res.append((TLV.kTLVType_PublicKey, SrpServer.to_byte_array(public_key),))
            d_res.append((TLV.kTLVType_Salt, SrpServer.to_byte_array(salt),))
            self._send_response_tlv(d_res)

            # store session
            self.server.sessions[self.session_id]['srp'] = server
            self.log_message('after step #2:\n%s', TLV.to_string(d_res))
            return

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M3:
            # step #4 Accessory -> iOS Device SRP Verify Response
            self.log_message('Step #4 /pair-setup')

            # 1) use ios pub key to compute shared secret key
            ios_pub_key = int.from_bytes(d_req[1][1], "big")
            server = self.server.sessions[self.session_id]['srp']
            server.set_client_public_key(ios_pub_key)

            hkdf_inst = hkdf.Hkdf('Pair-Setup-Encrypt-Salt'.encode(), SrpServer.to_byte_array(server.get_session_key()),
                                  hash=hashlib.sha512)
            session_key = hkdf_inst.expand('Pair-Setup-Encrypt-Info'.encode(), 32)
            self.server.sessions[self.session_id]['session_key'] = session_key

            # 2) verify ios proof
            ios_proof = int.from_bytes(d_req[2][1], "big")
            if not server.verify_clients_proof(ios_proof):
                d_res.append((TLV.kTLVType_State, TLV.M4,))
                d_res.append((TLV.kTLVType_Error, TLV.kTLVError_Authentication,))

                self._send_response_tlv(d_res)
                self.log_error('error in step #4 %s %s', d_res, self.server.sessions)
                return
            else:
                self.log_message('ios proof was verified')

            # 3) generate accessory proof
            accessory_proof = server.get_proof(ios_proof)

            # 4) create response tlv
            d_res.append((TLV.kTLVType_State, TLV.M4,))
            d_res.append((TLV.kTLVType_Proof, SrpServer.to_byte_array(accessory_proof),))

            # 5) send response tlv
            self._send_response_tlv(d_res)

            self.log_message('after step #4:\n%s', TLV.to_string(d_res))
            return

        if d_req[0][0] == TLV.kTLVType_State and d_req[0][1] == TLV.M5:
            # step #6 Accessory -> iOS Device Exchange Response
            self.log_message('Step #6 /pair-setup')

            # 1) Verify the iOS device's authTag
            # done by chacha20_aead_decrypt

            # 2) decrypt and test
            encrypted_data = d_req[1][1]
            decrypted_data = chacha20_aead_decrypt(bytes(), self.server.sessions[self.session_id]['session_key'],
                                                   'PS-Msg05'.encode(), bytes([0, 0, 0, 0]),
                                                   encrypted_data)
            if decrypted_data is False:
                d_res.append((TLV.kTLVType_State, TLV.M6,))
                d_res.append((TLV.kTLVType_Error, TLV.kTLVError_Authentication,))

                self.send_error_reply(TLV.M6, TLV.kTLVError_Authentication)
                self.log_error('error in step #6 %s %s', d_res, self.server.sessions)
                return

            d_req_2 = TLV.decode_bytearray(decrypted_data)

            # 3) Derive ios_device_x
            shared_secret = self.server.sessions[self.session_id]['srp'].get_session_key()
            hkdf_inst = hkdf.Hkdf('Pair-Setup-Controller-Sign-Salt'.encode(), SrpServer.to_byte_array(shared_secret),
                                  hash=hashlib.sha512)
            ios_device_x = hkdf_inst.expand('Pair-Setup-Controller-Sign-Info'.encode(), 32)

            # 4) construct ios_device_info
            ios_device_pairing_id = d_req_2[0][1]  # should be TLV.kTLVType_Identifier
            ios_device_ltpk = d_req_2[1][1]  # should be TLV.kTLVType_PublicKey
            ios_device_info = ios_device_x + ios_device_pairing_id + ios_device_ltpk

            # 5) verify signature
            ios_device_sig = d_req_2[2][1]  # should be [TLV.kTLVType_Signature

            verify_key = ed25519.VerifyingKey(bytes(ios_device_ltpk))
            try:
                verify_key.verify(bytes(ios_device_sig), bytes(ios_device_info))
            except ed25519.BadSignatureError:
                self.send_error_reply(TLV.M6, TLV.kTLVError_Authentication)
                self.log_error('error in step #6 %s %s', d_res, self.server.sessions)
                return

            # 6) save ios_device_pairing_id and ios_device_ltpk
            self.server.data.add_peer(ios_device_pairing_id, ios_device_ltpk, True)

            # Response Generation
            # 1) generate accessoryLTPK if not existing
            if self.server.data.accessory_ltsk is None or self.server.data.accessory_ltpk is None:
                accessory_ltsk, accessory_ltpk = ed25519.create_keypair()
                self.server.data.set_accessory_keys(
                    accessory_ltpk.to_bytes(),
                    accessory_ltsk.to_bytes(),
                )
            else:
                accessory_ltsk = ed25519.SigningKey(
                    self.server.data.accessory_ltsk + self.server.data.accessory_ltpk
                )
                accessory_ltpk = ed25519.VerifyingKey(self.server.data.accessory_ltpk)

            # 2) derive AccessoryX
            hkdf_inst = hkdf.Hkdf('Pair-Setup-Accessory-Sign-Salt'.encode(), SrpServer.to_byte_array(shared_secret),
                                  hash=hashlib.sha512)
            accessory_x = hkdf_inst.expand('Pair-Setup-Accessory-Sign-Info'.encode(), 32)

            # 3)
            accessory_info = accessory_x + self.server.data.accessory_pairing_id_bytes + accessory_ltpk.to_bytes()

            # 4) generate signature
            accessory_signature = accessory_ltsk.sign(accessory_info)

            # 5) construct sub_tlv
            sub_tlv = [
                (TLV.kTLVType_Identifier, self.server.data.accessory_pairing_id_bytes),
                (TLV.kTLVType_PublicKey, accessory_ltpk.to_bytes()),
                (TLV.kTLVType_Signature, accessory_signature)
            ]
            sub_tlv_b = TLV.encode_list(sub_tlv)

            # 6) encrypt sub_tlv
            encrypted_data_with_auth_tag = chacha20_aead_encrypt(bytes(),
                                                                 self.server.sessions[self.session_id]['session_key'],
                                                                 'PS-Msg06'.encode(),
                                                                 bytes([0, 0, 0, 0]),
                                                                 sub_tlv_b)
            tmp = bytearray(encrypted_data_with_auth_tag[0])
            tmp += encrypted_data_with_auth_tag[1]

            # 7) send response
            self.server.publish_device()
            d_res = [
                (TLV.kTLVType_State, TLV.M6,),
                (TLV.kTLVType_EncryptedData, tmp,)
            ]

            self._send_response_tlv(d_res)
            self.log_message('after step #6:\n%s', TLV.to_string(d_res))
            return

        self.send_error(HttpStatusCodes.METHOD_NOT_ALLOWED)

    def _send_response_tlv(self, d_res, close=False, status=HttpStatusCodes.OK):
        result_bytes = TLV.encode_list(d_res)

        self.send_response(status)
        # Send headers
        self.send_header('Content-Length', len(result_bytes))
        self.send_header('Content-Type', 'application/pairing+tlv8')
        self.send_header('Connection', 'keep-alive')
        self.end_headers()

        self.wfile.write(result_bytes)

    class Wrapper:
        """
        Wraps a bytes or byte array data into a file like object.
        """

        def __init__(self, data):
            self.data = data

        def makefile(self, arg):
            return io.BytesIO(self.data)

    def do_GET(self):
        """
        Can use
            * command
            * headers
            * path
            * ...
        :return:
        """
        absolute_path = self.path.split('?')[0]
        if absolute_path in self.PATHMAPPING:
            if 'GET' in self.PATHMAPPING[absolute_path]:
                # self.log_message('-' * 80 + '\ndo_GET / path: %s', self.path)
                self.PATHMAPPING[absolute_path]['GET']()
                return
        self.log_error('send error because of unmapped path: %s', self.path)
        self.send_error(HttpStatusCodes.NOT_FOUND)

    def do_POST(self):
        # read the body identified by its length
        content_length = int(self.headers['Content-Length'])
        self.body = self.rfile.read(content_length)
        if self.path in self.PATHMAPPING:
            if 'POST' in self.PATHMAPPING[self.path]:
                # self.log_message('-' * 80 + '\ndo_POST / path: %s', self.path)
                self.PATHMAPPING[self.path]['POST']()
                return
        self.log_error('send error because of unmapped path: %s', self.path)
        self.send_error(HttpStatusCodes.NOT_FOUND)

    def do_PUT(self):
        # read the body identified by its length
        content_length = int(self.headers['Content-Length'])
        self.body = self.rfile.read(content_length)
        if self.path in self.PATHMAPPING:
            if 'PUT' in self.PATHMAPPING[self.path]:
                # self.log_message('-' * 80 + '\ndo_PUT / path: %s', self.path)
                self.PATHMAPPING[self.path]['PUT']()
                return
        self.log_error('send error because of unmapped path: %s', self.path)
        self.send_error(HttpStatusCodes.NOT_FOUND)

    def log_message(self, format, *args):
        if self.server.logger is None:
            pass
        elif self.server.logger == sys.stderr:
            BaseHTTPRequestHandler.log_message(self, format, *args)
        else:
            self.server.logger.info("%s" % (format % args))

    def log_debug(self, format, *args):
        if self.server.logger is None:
            pass
        elif self.server.logger == sys.stderr:
            BaseHTTPRequestHandler.log_message(self, format, *args)
        else:
            self.server.logger.debug("%s" % (format % args))

    def log_error(self, format, *args):
        if self.server.logger is None:
            pass
        elif self.server.logger == sys.stderr:
            BaseHTTPRequestHandler.log_error(self, format, *args)
        else:
            self.server.logger.error("%s" % (format % args))



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class CharacteristicPermissions(object):
    """
    See table 5-4 page 67
    """
    paired_read = 'pr'
    paired_write = 'pw'
    events = 'ev'
    addition_authorization = 'aa'
    timed_write = 'tw'
    hidden = 'hd'



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import os


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit initialize controller storage app')
    parser.add_argument('-f', action='store', required=True, dest='file', help='HomeKit pairing data file')
    return parser.parse_args()


if __name__ == '__main__':
    args = setup_args_parser()
    if not os.path.isfile(args.file):
        try:
            with open(args.file, 'w') as fp:
                fp.write('{}')
        except PermissionError:
            print('Permission denied to create file "{f}".'.format(f=args.file))
    else:
        print('File "{f}" already exists.'.format(f=args.file))



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import sys
import logging

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit get_events app - listens to events from accessories')
    parser.add_argument('-f', action='store', required=True, dest='file', help='File with the pairing data')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    parser.add_argument('-c', action='append', required=True, dest='characteristics',
                        help='Use aid.iid value to change the value. Repeat to change multiple characteristics.')
    parser.add_argument('-e', action='store', required=False, dest='eventCount', help='max number of events before end',
                        default=-1, type=int)
    parser.add_argument('-s', action='store', required=False, dest='secondsCount', default=-1, type=int,
                        help='max number of seconds before end')
    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')
    add_log_arguments(parser)
    return parser.parse_args()


def func(events):
    for event in events:
        print('event for {aid}.{iid}: {event}'.format(aid=event[0], iid=event[1], event=event[2]))


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller(args.adapter)
    try:
        controller.load_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    if args.alias not in controller.get_pairings():
        print('"{a}" is no known alias'.format(a=args.alias))
        sys.exit(-1)

    try:
        pairing = controller.get_pairings()[args.alias]
        characteristics = [(int(c.split('.')[0]), int(c.split('.')[1])) for c in args.characteristics]
        results = pairing.get_events(characteristics, func, max_events=args.eventCount, max_seconds=args.secondsCount)
    except KeyboardInterrupt:
        sys.exit(-1)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    for key, value in results.items():
        aid = key[0]
        iid = key[1]
        status = value['status']
        desc = value['description']
        if status < 0:
            print('put_characteristics failed on {aid}.{iid} because: {reason} ({code})'.format(aid=aid, iid=iid,
                                                                                                reason=desc,
                                                                                                code=status))



# -*- coding: UTF-8 -*-

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Implements the Secure Remote Password (SRP) algorithm. More information can be found on
https://tools.ietf.org/html/rfc5054. See HomeKit spec page 36 for adjustments imposed by Apple.
"""
import crypt
import math
import hashlib


class Srp:
    def __init__(self):
        # generator as defined by 3072bit group of RFC 5054
        self.g = int(b'5', 16)
        # modulus as defined by 3072bit group of RFC 5054
        self.n = int(b'''\
FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E08\
8A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B\
302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9\
A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE6\
49286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8\
FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D\
670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C\
180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF695581718\
3995497CEA956AE515D2261898FA051015728E5A8AAAC42DAD33170D\
04507A33A85521ABDF1CBA64ECFB850458DBEF0A8AEA71575D060C7D\
B3970F85A6E1E4C7ABF5AE8CDB0933D71E8C94E04A25619DCEE3D226\
1AD2EE6BF12FFA06D98A0864D87602733EC86A64521F2B18177B200C\
BBE117577A615D6C770988C0BAD946E208E24FA074E5AB3143DB5BFC\
E0FD108E4B82D120A93AD2CAFFFFFFFFFFFFFFFF''', 16)
        # HomeKit requires SHA-512 (See page 36)
        self.h = hashlib.sha512
        self.A = None
        self.B = None
        self.salt = None
        self.username = None
        self.password = None

    @staticmethod
    def generate_private_key():
        """
        Static function to generate a 16 byte random key.

        :return: the key as an integer
        """
        private_key = crypt.mksalt(crypt.METHOD_SHA512)[3:].encode()
        return int.from_bytes(private_key, "big")

    def _calculate_k(self) -> int:
        # calculate k (see https://tools.ietf.org/html/rfc5054#section-2.5.3)
        hash_instance = self.h()
        n = Srp.to_byte_array(self.n)
        g = bytearray.fromhex((383 * '00' + '05'))  # 383 * b'0' + '5'.encode()
        hash_instance.update(n)
        hash_instance.update(g)
        k = int.from_bytes(hash_instance.digest(), "big")
        return k

    def _calculate_u(self) -> int:
        if self.A is None:
            raise RuntimeError('Client\'s public key is missing')
        if self.B is None:
            raise RuntimeError('Server\'s public key is missing')
        hash_instance = self.h()
        A_b = Srp.to_byte_array(self.A)
        B_b = Srp.to_byte_array(self.B)
        hash_instance.update(A_b)
        hash_instance.update(B_b)
        u = int.from_bytes(hash_instance.digest(), "big")
        return u

    def get_session_key(self) -> int:
        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.get_shared_secret()))
        hash_value = int.from_bytes(hash_instance.digest(), "big")
        return hash_value

    @staticmethod
    def to_byte_array(num: int) -> bytearray:
        return bytearray(num.to_bytes(int(math.ceil(num.bit_length() / 8)), "big"))

    def _calculate_x(self) -> int:
        i = (self.username + ':' + self.password).encode()
        hash_instance = self.h()
        hash_instance.update(i)
        hash_value = hash_instance.digest()

        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.salt))
        hash_instance.update(hash_value)

        return int.from_bytes(hash_instance.digest(), "big")

    def get_shared_secret(self):
        raise NotImplementedError()


class SrpClient(Srp):
    """
    Implements all functions that are required to simulate an iOS HomeKit controller
    """
    def __init__(self, username: str, password: str):
        Srp.__init__(self)
        self.username = username
        self.password = password
        self.salt = None
        self.a = self.generate_private_key()
        self.A = pow(self.g, self.a, self.n)
        self.B = None

    def set_salt(self, salt):
        if isinstance(salt, bytearray):
            self.salt = int.from_bytes(salt, "big")
        else:
            self.salt = salt

    def get_public_key(self):
        return pow(self.g, self.a, self.n)

    def set_server_public_key(self, B):
        if isinstance(B, bytearray):
            self.B = int.from_bytes(B, "big")
        else:
            self.B = B

    def get_shared_secret(self):
        if self.B is None:
            raise RuntimeError('Server\'s public key is missing')
        u = self._calculate_u()
        x = self._calculate_x()
        k = self._calculate_k()
        tmp1 = (self.B - (k * pow(self.g, x, self.n)))
        tmp2 = (self.a + (u * x))  # % self.n
        S = pow(tmp1, tmp2, self.n)
        return S

    def get_proof(self):
        if self.B is None:
            raise RuntimeError('Server\'s public key is missing')

        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.n))
        hN = bytearray(hash_instance.digest())

        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.g))
        hg = bytearray(hash_instance.digest())

        for index in range(0, len(hN)):
            hN[index] ^= hg[index]

        u = self.username.encode()
        hash_instance = self.h()
        hash_instance.update(u)
        hu = hash_instance.digest()
        K = Srp.to_byte_array(self.get_session_key())

        hash_instance = self.h()
        hash_instance.update(hN)
        hash_instance.update(hu)
        hash_instance.update(Srp.to_byte_array(self.salt))
        hash_instance.update(Srp.to_byte_array(self.A))
        hash_instance.update(Srp.to_byte_array(self.B))
        hash_instance.update(K)
        return int.from_bytes(hash_instance.digest(), "big")

    def verify_servers_proof(self, M):
        if isinstance(M, bytearray):
            tmp = int.from_bytes(M, "big")
        else:
            tmp = M
        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.A))
        hash_instance.update(Srp.to_byte_array(self.get_proof()))
        hash_instance.update(Srp.to_byte_array(self.get_session_key()))
        return tmp == int.from_bytes(hash_instance.digest(), "big")


class SrpServer(Srp):
    """
    Implements all functions that are required to simulate an iOS HomeKit accessory
    """
    def __init__(self, username, password):
        Srp.__init__(self)
        self.username = username
        self.salt = SrpServer._create_salt()
        self.password = password
        self.verifier = self._get_verifier()
        self.b = self.generate_private_key()
        k = self._calculate_k()
        g_b = pow(self.g, self.b, self.n)
        self.B = (k * self.verifier + g_b) % self.n
        self.A = None

    @staticmethod
    def _create_salt() -> int:
        # generate random salt
        salt = crypt.mksalt(crypt.METHOD_SHA512)[3:]
        assert len(salt) == 16
        salt_b = salt.encode()
        return int.from_bytes(salt_b, "big")

    def _get_verifier(self) -> int:
        hash_value = self._calculate_x()
        v = pow(self.g, hash_value, self.n)
        return v

    def set_client_public_key(self, A):
        self.A = A

    def get_salt(self):
        return self.salt

    def get_public_key(self):
        k = self._calculate_k()
        return (k * self.verifier + pow(self.g, self.b, self.n)) % self.n

    def get_shared_secret(self):
        if self.A is None:
            raise RuntimeError('Client\'s public key is missing')

        tmp1 = self.A * pow(self.verifier, self._calculate_u(), self.n)
        return pow(tmp1, self.b, self.n)

    def verify_clients_proof(self, m) -> bool:
        if self.B is None:
            raise RuntimeError('Server\'s public key is missing')

        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.n))
        hN = bytearray(hash_instance.digest())

        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.g))
        hg = bytearray(hash_instance.digest())

        for index in range(0, len(hN)):
            hN[index] ^= hg[index]

        u = self.username.encode()
        hash_instance = self.h()
        hash_instance.update(u)
        hu = hash_instance.digest()
        K = Srp.to_byte_array(self.get_session_key())

        hash_instance = self.h()
        hash_instance.update(hN)
        hash_instance.update(hu)
        hash_instance.update(Srp.to_byte_array(self.salt))
        hash_instance.update(Srp.to_byte_array(self.A))
        hash_instance.update(Srp.to_byte_array(self.B))
        hash_instance.update(K)
        return m == int.from_bytes(hash_instance.digest(), "big")

    def get_proof(self, m) -> int:
        hash_instance = self.h()
        hash_instance.update(Srp.to_byte_array(self.A))
        hash_instance.update(Srp.to_byte_array(m))
        hash_instance.update(Srp.to_byte_array(self.get_session_key()))
        return int.from_bytes(hash_instance.digest(), "big")



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'Device', 'BlePairing', 'BleSession', 'find_characteristic_by_uuid', 'create_ble_pair_setup_write'
]

import time
import logging
import random
import sys
import uuid
import struct
from distutils.util import strtobool

from homekit.controller.tools import AbstractPairing
from homekit.protocol.tlv import TLV
from homekit.model.characteristics import CharacteristicsTypes
from homekit.protocol import get_session_keys
from homekit.protocol.opcodes import HapBleOpCodes
from homekit.protocol.statuscodes import HapBleStatusCodes
from homekit.model.services.service_types import ServicesTypes
from homekit.crypto import chacha20_aead_decrypt, chacha20_aead_encrypt
from homekit.model.characteristics.characteristic_formats import BleCharacteristicFormats, CharacteristicFormats
from homekit.model.characteristics.characteristic_units import BleCharacteristicUnits
from homekit.exceptions import FormatError, RequestRejected, AccessoryDisconnectedError

from homekit.tools import BLE_TRANSPORT_SUPPORTED

if BLE_TRANSPORT_SUPPORTED:
    from .device import DeviceManager, Device
else:
    # this empty class is required to be inherited by ResolvingManager
    class DeviceManager:
        pass

# the uuid of the ble descriptors that hold the characteristic instance id as value (see page 128)
CharacteristicInstanceID = 'dc46f0fe-81d2-4616-b5d9-6abdd796939a'

logger = logging.getLogger(__name__)


class BlePairing(AbstractPairing):

    def __init__(self, pairing_data, adapter='hci0'):
        """
        Initialize a Pairing by using the data either loaded from file or obtained after calling
        Controller.perform_pairing().

        :param pairing_data:
        """
        self.adapter = adapter
        self.pairing_data = pairing_data
        self.session = None

    def close(self):
        pass

    def list_accessories_and_characteristics(self):
        if 'accessories' in self.pairing_data:
            return self.pairing_data['accessories']

        manager = DeviceManager(self.adapter)
        device = manager.make_device(self.pairing_data['AccessoryMAC'])
        device.connect()
        resolved_data = read_characteristics(device)
        self.pairing_data['accessories'] = resolved_data['data']
        return resolved_data['data']

    def list_pairings(self):
        if not self.session:
            self.session = BleSession(self.pairing_data, self.adapter)
        request_tlv = TLV.encode_list([
            (TLV.kTLVType_State, TLV.M1),
            (TLV.kTLVType_Method, TLV.ListPairings)
        ])
        request_tlv = TLV.encode_list([
            (TLV.kTLVHAPParamParamReturnResponse, bytearray(b'\x01')),
            (TLV.kTLVHAPParamValue, request_tlv)
        ])
        body = len(request_tlv).to_bytes(length=2, byteorder='little') + request_tlv

        cid = -1
        for a in self.pairing_data['accessories']:
            for s in a['services']:
                for c in s['characteristics']:
                    if CharacteristicsTypes.get_short_uuid(c['type'].upper()) == CharacteristicsTypes.PAIRING_PAIRINGS:
                        cid = c['iid']
        fc, _ = self.session.find_characteristic_by_iid(cid)
        response = self.session.request(fc, cid, HapBleOpCodes.CHAR_WRITE, body)
        response = TLV.decode_bytes(response[1])
        tmp = []
        r = {}
        for d in response[1:]:
            if d[0] == TLV.kTLVType_Identifier:
                r = {}
                tmp.append(r)
                r['pairingId'] = d[1].decode()
            if d[0] == TLV.kTLVType_PublicKey:
                r['publicKey'] = d[1].hex()
            if d[0] == TLV.kTLVType_Permissions:
                controller_type = 'regular'
                if d[1] == b'\x01':
                    controller_type = 'admin'
                r['permissions'] = int.from_bytes(d[1], byteorder='little')
                r['controllerType'] = controller_type
        return tmp

    def get_events(self, characteristics, callback_fun, max_events=-1, max_seconds=-1):
        # TODO implementation still missing
        pass

    def identify(self):
        """
        This call can be used to trigger the identification of a paired accessory. A successful call should
        cause the accessory to perform some specific action by which it can be distinguished from the others (blink a
        LED for example).

        It uses the identify characteristic as described on page 152 of the spec.

        :return True, if the identification was run, False otherwise
        """
        if not self.session:
            self.session = BleSession(self.pairing_data, self.adapter)
        cid = -1
        aid = -1
        for a in self.pairing_data['accessories']:
            for s in a['services']:
                for c in s['characteristics']:
                    if CharacteristicsTypes.get_short_uuid(c['type'].upper()) == CharacteristicsTypes.IDENTIFY:
                        aid = a['aid']
                        cid = c['iid']
        self.put_characteristics([(aid, cid, True)])
        # TODO check for errors
        return True

    def get_characteristics(self, characteristics, include_meta=False, include_perms=False, include_type=False,
                            include_events=False):
        """
        This method is used to get the current readouts of any characteristic of the accessory.

        :param characteristics: a list of 2-tupels of accessory id and instance id
        :param include_meta: if True, include meta information about the characteristics. This contains the format and
                             the various constraints like maxLen and so on.
        :param include_perms: if True, include the permissions for the requested characteristics.
        :param include_type: if True, include the type of the characteristics in the result. See CharacteristicsTypes
                             for translations.
        :param include_events: if True on a characteristics that supports events, the result will contain information if
                               the controller currently is receiving events for that characteristic. Key is 'ev'.
        :return: a dict mapping 2-tupels of aid and iid to dicts with value or status and description, e.g.
                 {(1, 8): {'value': 23.42}
                  (1, 37): {'description': 'Resource does not exist.', 'status': -70409}
                 }
        """
        if not self.session:
            self.session = BleSession(self.pairing_data, self.adapter)

        results = {}
        for aid, cid in characteristics:
            try:
                fc, fc_info = self.session.find_characteristic_by_iid(cid)
                if not fc or not fc_info:
                    results[(aid, cid)] = {
                        'status': HapBleStatusCodes.INVALID_REQUEST,
                        'description': HapBleStatusCodes[HapBleStatusCodes.INVALID_REQUEST]
                    }
                    continue
                response = self.session.request(fc, cid, HapBleOpCodes.CHAR_READ)
            except Exception as e:
                self.session.close()
                self.session = None
                raise e

            value = self._convert_to_python(aid, cid, response[1]) if 1 in response else None

            results[(aid, cid)] = {
                'value': value
            }

        return results

    def _convert_from_python(self, aid, cid, value):
        """
        Convert a python value into a representation usable for the Bluetooth LE transport. The type to be used will be
        queried from the characteristic.

        :param aid: the accessory id
        :param cid: the characteristic id
        :param value: the value to convert (as python value)
        :return: the converted value as bytes
        """
        char_format = None
        characteristic = self._find_characteristic_in_pairing_data(aid, cid)
        if characteristic:
            char_format = characteristic['format']
        logger.debug('value: %s format: %s', value, char_format)

        if char_format == CharacteristicFormats.bool:
            try:
                val = strtobool(str(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
            return struct.pack('?', val)

        # no  more boolean input after this line
        if type(value) == bool:
            raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))

        if char_format == CharacteristicFormats.float:
            try:
                return struct.pack('f', float(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))

        if char_format == CharacteristicFormats.string:
            if type(value) != str:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
            return value.encode('UTF-8')

        if char_format == CharacteristicFormats.data or char_format == CharacteristicFormats.tlv8:
            return value.hex().encode()

        # from here only integer values of different sizes
        if char_format == CharacteristicFormats.int:
            try:
                if float(value) != int(value):
                    raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
                value = struct.pack('i', int(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
        elif char_format == CharacteristicFormats.uint8:
            try:
                if float(value) != int(value):
                    raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
                value = struct.pack('B', int(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
            except struct.error:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
        elif char_format == CharacteristicFormats.uint16:
            try:
                if float(value) != int(value):
                    raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
                value = struct.pack('H', int(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
            except struct.error:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
        elif char_format == CharacteristicFormats.uint32:
            try:
                if float(value) != int(value):
                    raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
                value = struct.pack('I', int(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
            except struct.error:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
        elif char_format == CharacteristicFormats.uint64:
            try:
                if float(value) != int(value):
                    raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
                value = struct.pack('Q', int(value))
            except ValueError:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
            except struct.error:
                raise FormatError('"{v}" is no valid "{t}"!'.format(v=value, t=char_format))
        return value

    def _convert_to_python(self, aid, cid, value):
        """
        Convert a value from the Bluetooth LE transport to python values.

        :param aid: the accessory id
        :param cid: the characteristic id
        :param value: the value as bytes
        :return: the converted value as python value
        """
        char_format = None
        characteristic = self._find_characteristic_in_pairing_data(aid, cid)
        if characteristic:
            char_format = characteristic['format']
        logger.debug('value: %s format: %s', value.hex(), char_format)

        if char_format == CharacteristicFormats.bool:
            value = struct.unpack('?', value)[0]
        elif char_format == CharacteristicFormats.uint8:
            value = struct.unpack('B', value)[0]
        elif char_format == CharacteristicFormats.uint16:
            value = struct.unpack('H', value)[0]
        elif char_format == CharacteristicFormats.uint32:
            value = struct.unpack('I', value)[0]
        elif char_format == CharacteristicFormats.uint64:
            value = struct.unpack('Q', value)[0]
        elif char_format == CharacteristicFormats.int:
            value = struct.unpack('i', value)[0]
        elif char_format == CharacteristicFormats.float:
            value = struct.unpack('f', value)[0]
        elif char_format == CharacteristicFormats.string:
            value = value.decode('UTF-8')
        else:
            value = value.hex()
        return value

    def _find_characteristic_in_pairing_data(self, aid, cid):
        """
        Iterate over the accessories in the pairing data. If a characteristic with the correct accessory id and
        characteristic id is found, the characteristic is returned. If no characteristic is found, it returns None.

        :param aid: the accessory id
        :param cid: the characteristic id
        :return: the characteristic or None
        """
        for a in self.pairing_data['accessories']:
            if a['aid'] == int(aid):
                for s in a['services']:
                    for c in s['characteristics']:
                        if c['iid'] == int(cid):
                            return c
        return None

    def put_characteristics(self, characteristics, do_conversion=False):
        """
        Update the values of writable characteristics. The characteristics have to be identified by accessory id (aid),
        instance id (iid). If do_conversion is False (the default), the value must be of proper format for the
        characteristic since no conversion is done. If do_conversion is True, the value is converted.

        :param characteristics: a list of 3-tupels of accessory id, instance id and the value
        :param do_conversion: select if conversion is done (False is default)
        :return: a dict from (aid, iid) onto {status, description}
        :raises FormatError: if the input value could not be converted to the target type and conversion was
                             requested
        """
        if not self.session:
            self.session = BleSession(self.pairing_data, self.adapter)

        results = {}

        for aid, cid, value in characteristics:
            # reply with an error if the characteristic does not exist
            if not self._find_characteristic_in_pairing_data(aid, cid):
                results[(aid, cid)] = {
                    'status': HapBleStatusCodes.INVALID_REQUEST,
                    'description': HapBleStatusCodes[HapBleStatusCodes.INVALID_REQUEST]
                }
                continue

            value = TLV.encode_list([(1, self._convert_from_python(aid, cid, value))])
            body = len(value).to_bytes(length=2, byteorder='little') + value

            try:
                fc, fc_info = self.session.find_characteristic_by_iid(cid)
                response = self.session.request(fc, cid, HapBleOpCodes.CHAR_WRITE, body)
                logger.debug('response %s', response)
                # TODO does the response contain useful information here?
            except RequestRejected as e:
                results[(aid, cid)] = {
                    'status': e.status,
                    'description': e.message,
                }
            except Exception as e:
                self.session.close()
                self.session = None
                raise e

        return results

    def add_pairing(self, additional_controller_pairing_identifier, ios_device_ltpk, permissions):
        if not self.session:
            self.session = BleSession(self.pairing_data, self.adapter)
        if permissions == 'User':
            permissions = TLV.kTLVType_Permission_RegularUser
        elif permissions == 'Admin':
            permissions = TLV.kTLVType_Permission_AdminUser
        else:
            print('UNKNOWN')

        request_tlv = TLV.encode_list([
            (TLV.kTLVType_State, TLV.M1),
            (TLV.kTLVType_Method, TLV.AddPairing),
            (TLV.kTLVType_Identifier, additional_controller_pairing_identifier.encode()),
            (TLV.kTLVType_PublicKey, bytes.fromhex(ios_device_ltpk)),
            (TLV.kTLVType_Permissions, permissions)
        ])

        request_tlv = TLV.encode_list([
            (TLV.kTLVHAPParamParamReturnResponse, bytearray(b'\x01')),
            (TLV.kTLVHAPParamValue, request_tlv)
        ])
        body = len(request_tlv).to_bytes(length=2, byteorder='little') + request_tlv

        cid = -1
        for a in self.pairing_data['accessories']:
            for s in a['services']:
                for c in s['characteristics']:
                    if CharacteristicsTypes.get_short_uuid(c['type'].upper()) == CharacteristicsTypes.PAIRING_PAIRINGS:
                        cid = c['iid']
        fc, _ = self.session.find_characteristic_by_iid(cid)
        response = self.session.request(fc, cid, HapBleOpCodes.CHAR_WRITE, body)
        # TODO handle response properly
        print('unhandled response:', response)


class BleSession(object):

    def __init__(self, pairing_data, adapter):
        self.adapter = adapter
        self.pairing_data = pairing_data
        self.c2a_counter = 0
        self.a2c_counter = 0
        self.c2a_key = None
        self.a2c_key = None
        self.device = None
        mac_address = self.pairing_data['AccessoryMAC']

        manager = DeviceManager(self.adapter)

        self.device = manager.make_device(mac_address)
        logger.debug('connecting to device')
        self.device.connect()
        logger.debug('connected to device')

        uuid_map = {}
        for s in self.device.services:
            for c in s.characteristics:
                uuid_map[(s.uuid.upper(), c.uuid.upper())] = c

        self.uuid_map = {}
        self.iid_map = {}
        self.short_map = {}
        for a in pairing_data['accessories']:
            for s in a['services']:
                s_short = None
                if s['type'].endswith(ServicesTypes.baseUUID):
                    s_short = s['type'].split('-', 1)[0].lstrip('0')

                for c in s['characteristics']:
                    char = uuid_map.get((s['type'], c['type']), None)
                    if not char:
                        continue
                    self.iid_map[c['iid']] = (char, c)
                    self.uuid_map[(s['type'], c['type'])] = (char, c)

                    if s_short and c['type'].endswith(CharacteristicsTypes.baseUUID):
                        c_short = c['type'].split('-', 1)[0].lstrip('0')
                        self.short_map[(s_short, c_short)] = (char, c)

                    service_type_short = ServicesTypes.get_short(s['type'])
                    characteristic_type_short = CharacteristicsTypes.get_short(c['type'])
                    self.short_map[service_type_short, characteristic_type_short] = (char, c)

        pair_verify_char, pair_verify_char_info = self.short_map.get(
            (ServicesTypes.PAIRING_SERVICE, CharacteristicsTypes.PAIR_VERIFY),
            (None, None)
        )

        if not pair_verify_char:
            print('verify characteristic not found')
            # TODO Have exception here
            sys.exit(-1)

        write_fun = create_ble_pair_setup_write(pair_verify_char, pair_verify_char_info['iid'])
        self.c2a_key, self.a2c_key = get_session_keys(None, self.pairing_data, write_fun)
        logger.debug('pair_verified, keys: \n\t\tc2a: %s\n\t\ta2c: %s', self.c2a_key.hex(), self.a2c_key.hex())

        self.c2a_counter = 0
        self.a2c_counter = 0

    def __del__(self):
        self.close()

    def close(self):
        logger.debug('closing session')
        self.device.disconnect()

    def find_characteristic_by_iid(self, cid):
        return self.iid_map.get(cid, (None, None))

    def request(self, feature_char, feature_char_id, op, body=None):
        transaction_id = random.randrange(0, 255)

        data = bytearray([0x00, op, transaction_id])
        data.extend(feature_char_id.to_bytes(length=2, byteorder='little'))

        if body:
            logger.debug('body: %s', body)
            data.extend(body)

        logger.debug('data: %s', data)

        cnt_bytes = self.c2a_counter.to_bytes(8, byteorder='little')
        cipher_and_mac = chacha20_aead_encrypt(bytes(), self.c2a_key, cnt_bytes, bytes([0, 0, 0, 0]), data)
        cipher_and_mac[0].extend(cipher_and_mac[1])
        data = cipher_and_mac[0]
        logger.debug('cipher and mac %s', cipher_and_mac[0].hex())

        result = feature_char.write_value(value=data)
        logger.debug('write resulted in: %s', result)

        self.c2a_counter += 1

        data = []
        while not data or len(data) == 0:
            time.sleep(1)
            logger.debug('reading characteristic')
            data = feature_char.read_value()
            if not data and not self.device.is_connected():
                raise AccessoryDisconnectedError('Characteristic read failed')

        resp_data = bytearray([b for b in data])
        logger.debug('read: %s', bytearray(resp_data).hex())

        data = chacha20_aead_decrypt(bytes(), self.a2c_key, self.a2c_counter.to_bytes(8, byteorder='little'),
                                     bytes([0, 0, 0, 0]), resp_data)

        logger.debug('decrypted: %s', bytearray(data).hex())

        if not data:
            return {}

        # parse header and check stuff
        logger.debug('parse sig read response %s', bytes([int(a) for a in data]).hex())

        # handle the header data
        cf = data[0]
        logger.debug('control field %d', cf)
        tid = data[1]
        logger.debug('transaction id %d (expected was %d)', tid, transaction_id)
        status = data[2]
        logger.debug('status code %d (%s)', status, HapBleStatusCodes[status])
        assert cf == 0x02
        assert tid == transaction_id

        if status != HapBleStatusCodes.SUCCESS:
            raise RequestRejected(status, HapBleStatusCodes[status])

        self.a2c_counter += 1

        # get body length
        length = int.from_bytes(data[3:5], byteorder='little')
        logger.debug('expected body length %d (got %d)', length, len(data[5:]))

        # parse tlvs and analyse information
        tlv = TLV.decode_bytes(data[5:])
        logger.debug('received TLV: %s', TLV.to_string(tlv))
        return dict(tlv)


def find_characteristic_by_uuid(device, service_uuid, char_uuid):
    """
    # TODO document me

    :param device:
    :param service_uuid:
    :param char_uuid:
    :return:
    """
    service_found = None
    logger.debug('services: %s', device.services)
    for possible_service in device.services:
        if ServicesTypes.get_short(service_uuid.upper()) == ServicesTypes.get_short(possible_service.uuid.upper()):
            service_found = possible_service
            break
    logger.debug('searched service: %s', service_found)

    if not service_found:
        logging.error('searched service not found.')
        return None, None

    result_char = None
    result_char_id = None
    for characteristic in service_found.characteristics:
        logger.debug('char: %s %s', characteristic.uuid, CharacteristicsTypes.get_short(characteristic.uuid.upper()))
        if CharacteristicsTypes.get_short(char_uuid.upper()) == CharacteristicsTypes.get_short(
                characteristic.uuid.upper()):
            result_char = characteristic
            for descriptor in characteristic.descriptors:
                value = descriptor.read_value()
                if descriptor.uuid == CharacteristicInstanceID:
                    cid = int.from_bytes(value, byteorder='little')
                    result_char_id = cid

    if not result_char:
        logging.error('searched char not found.')
        return None, None

    logger.debug('searched char: %s %s', result_char, result_char_id)
    return result_char, result_char_id


def create_ble_pair_setup_write(characteristic, characteristic_id):
    def write(request, expected):
        # TODO document me
        logger.debug('entering write function %s', TLV.to_string(TLV.decode_bytes(request)))
        request_tlv = TLV.encode_list([
            (TLV.kTLVHAPParamParamReturnResponse, bytearray(b'\x01')),
            (TLV.kTLVHAPParamValue, request)
        ])
        transaction_id = random.randrange(0, 255)
        data = bytearray([0x00, HapBleOpCodes.CHAR_WRITE, transaction_id])
        data.extend(characteristic_id.to_bytes(length=2, byteorder='little'))
        data.extend(len(request_tlv).to_bytes(length=2, byteorder='little'))
        data.extend(request_tlv)
        logger.debug('sent %s', bytes(data).hex())
        characteristic.write_value(value=data)
        data = []
        while len(data) == 0:
            time.sleep(1)
            logger.debug('reading characteristic')
            data = characteristic.read_value()
        resp_data = [b for b in data]

        expected_length = int.from_bytes(bytes(resp_data[3:5]), byteorder='little')
        logger.debug(
            'control field: {c:x}, tid: {t:x}, status: {s:x}, length: {length}'.format(c=resp_data[0], t=resp_data[1],
                                                                                       s=resp_data[2],
                                                                                       length=expected_length))
        while len(resp_data[3:]) < expected_length:
            time.sleep(1)
            logger.debug('reading characteristic')
            data = characteristic.read_value()
            resp_data.extend([b for b in data])
            logger.debug('data %s of %s', len(resp_data[3:]), expected_length)

        logger.debug('received %s', bytes(resp_data).hex())
        logger.debug('decode %s', bytes(resp_data[5:]).hex())
        resp_tlv = TLV.decode_bytes(bytes([int(a) for a in resp_data[5:]]), expected=[TLV.kTLVHAPParamValue])
        result = TLV.decode_bytes(resp_tlv[0][1], expected)
        logger.debug('leaving write function %s', TLV.to_string(result))
        return result

    return write


class ResolvingManager(DeviceManager):
    """
    DeviceManager implementation that stops running after a given device was discovered.
    """

    def __init__(self, adapter_name, mac):
        self.mac = mac
        DeviceManager.__init__(self, adapter_name=adapter_name)

    def device_discovered(self, device):
        logger.debug('discovered %s', device.mac_address.upper())
        if device.mac_address.upper() == self.mac.upper():
            self.stop()
            # the searched device was found, so we can stop discovery here
            self.stop_discovery()


def read_characteristics(device):
    # TODO document me
    # FIXME: This only works on non secure sessions
    logger.debug('resolved %d services', len(device.services))

    a_data = {
        'aid': 1,
        'services': []
    }

    resolved_data = {
        'data': [a_data],
    }

    for service in device.services:
        logger.debug('found service with UUID %s (%s)', service.uuid,
                     ServicesTypes.get_short(service.uuid.upper()))

        s_data = {
            'characteristics': [
            ],
            'iid': None
        }

        for characteristic in service.characteristics:
            logger.debug('\tfound characteristic with UUID %s (%s)', characteristic.uuid,
                         CharacteristicsTypes.get_short(characteristic.uuid.upper()))

            if characteristic.uuid.upper() == CharacteristicsTypes.SERVICE_INSTANCE_ID:
                sid = int.from_bytes(characteristic.read_value(), byteorder='little')
                logger.debug('\t\tread service id %d', sid)
                s_data['iid'] = sid
            else:
                c_data = {
                    'iid': None,
                    'type': characteristic.uuid.upper(),
                    'perms': []
                }
                iid = None
                for descriptor in characteristic.descriptors:
                    value = descriptor.read_value()
                    if descriptor.uuid == CharacteristicInstanceID:
                        iid = int.from_bytes(value, byteorder='little')
                        logger.debug('\t\tread characteristic id %d', iid)
                        c_data['iid'] = iid
                    else:
                        # print('\t\t', 'D', descriptor.uuid, value)
                        pass

                if iid:
                    v = iid.to_bytes(length=2, byteorder='little')
                    tid = random.randrange(0, 255)
                    characteristic.write_value([0x00, HapBleOpCodes.CHAR_SIG_READ, tid, v[0], v[1]])
                    d = parse_sig_read_response(characteristic.read_value(), tid)
                    for k in d:
                        if k == 'service_type':
                            s_data['type'] = d[k].upper()
                        elif k == 'sid':
                            s_data['iid'] = d[k]
                        else:
                            c_data[k] = d[k]

                if c_data['iid']:
                    s_data['characteristics'].append(c_data)

        if s_data['iid']:
            a_data['services'].append(s_data)

    logger.debug('data: %s', resolved_data)

    return resolved_data


def parse_sig_read_response(data, expected_tid):
    # TODO document me
    # parse header and check stuff
    logger.debug('parse sig read response %s', bytes([int(a) for a in data]).hex())

    # handle the header data
    cf = data[0]
    logger.debug('control field %d', cf)
    tid = data[1]
    logger.debug('transaction id %d (expected was %d)', tid, expected_tid)
    status = data[2]
    logger.debug('status code %d (%s)', status, HapBleStatusCodes[status])
    assert cf == 0x02
    assert tid == expected_tid
    assert status == HapBleStatusCodes.SUCCESS

    # get body length
    length = int.from_bytes(data[3:5], byteorder='little')
    logger.debug('expected body length %d (got %d)', length, len(data[5:]))

    # parse tlvs and analyse information
    tlv = TLV.decode_bytes(data[5:])

    description = ''
    characteristic_format = ''
    characteristic_range = None
    characteristic_step = None
    for t in tlv:
        if t[0] == TLV.kTLVHAPParamCharacteristicType:
            chr_type = [int(a) for a in t[1]]
            chr_type.reverse()
            chr_type = str(uuid.UUID(''.join('%02x' % b for b in chr_type)))
        if t[0] == TLV.kTLVHAPParamServiceInstanceId:
            svc_id = int.from_bytes(t[1], byteorder='little')
        if t[0] == TLV.kTLVHAPParamServiceType:
            svc_type = [int(a) for a in t[1]]
            svc_type.reverse()
            svc_type = str(uuid.UUID(''.join('%02x' % b for b in svc_type)))
        if t[0] == TLV.kTLVHAPParamHAPCharacteristicPropertiesDescriptor:
            chr_prop_int = int.from_bytes(t[1], byteorder='little')
        if t[0] == TLV.kTLVHAPParamGATTUserDescriptionDescriptor:
            description = t[1].decode()
        if t[0] == TLV.kTLVHAPParamHAPValidValuesDescriptor:
            print('valid values', t[1])
        if t[0] == TLV.kTLVHAPParamHAPValidValuesRangeDescriptor:
            print('valid values range', t[1])
        if t[0] == TLV.kTLVHAPParamGATTPresentationFormatDescriptor:
            unit_bytes = t[1][2:4]
            unit_bytes.reverse()
            characteristic_format = BleCharacteristicFormats.get(int(t[1][0]), 'unknown')
            unit = BleCharacteristicUnits.get(int.from_bytes(unit_bytes, byteorder='big'), 'unknown')
        if t[0] == TLV.kTLVHAPParamGATTValidRange:
            logger.debug('range: %s', t[1].hex())
            lower = None
            upper = None
            if characteristic_format == 'int32' or characteristic_format == 'int':
                (lower, upper) = struct.unpack('ii', t[1])
            if characteristic_format == 'uint8':
                (lower, upper) = struct.unpack('BB', t[1])
            if characteristic_format == 'float':
                (lower, upper) = struct.unpack('ff', t[1])
            # TODO include all formats!
            characteristic_range = (lower, upper)
        if t[0] == TLV.kTLVHAPParamHAPStepValueDescriptor:
            characteristic_step = None
            if characteristic_format == 'int32':
                characteristic_step = struct.unpack('i', t[1])[0]
            if characteristic_format == 'uint8':
                characteristic_step = struct.unpack('B', t[1])[0]
            # TODO include all formats!

    # parse permissions
    # TODO refactor!
    perms = []
    if (chr_prop_int & 0x0001) > 0:
        perms.append('r')
    if (chr_prop_int & 0x0002) > 0:
        perms.append('w')
    if (chr_prop_int & 0x0004) > 0:
        perms.append('aad')
    if (chr_prop_int & 0x0008) > 0:
        perms.append('tw')
    if (chr_prop_int & 0x0010) > 0:
        perms.append('pr')
    if (chr_prop_int & 0x0020) > 0:
        perms.append('pw')
    if (chr_prop_int & 0x0040) > 0:
        perms.append('hd')
    if (chr_prop_int & 0x0080) > 0:
        perms.append('evc')
    if (chr_prop_int & 0x0100) > 0:
        perms.append('evd')

    result = {'description': description, 'perms': perms, 'format': characteristic_format, 'unit': unit,
              'range': characteristic_range, 'step': characteristic_step,
              'type': chr_type.upper(), 'sid': svc_id, 'service_type': svc_type}
    logger.debug('result: %s', str(result))

    return result



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging

logger = logging.getLogger('homekit.protocol.tlv')


class TLV:
    """
    as described in Appendix 12 (page 251)
    """

    # Steps
    M1 = bytearray(b'\x01')
    M2 = bytearray(b'\x02')
    M3 = bytearray(b'\x03')
    M4 = bytearray(b'\x04')
    M5 = bytearray(b'\x05')
    M6 = bytearray(b'\x06')

    # Methods (see table 4-4 page 60)
    PairSetup = bytearray(b'\x01')
    PairVerify = bytearray(b'\x02')
    AddPairing = bytearray(b'\x03')
    RemovePairing = bytearray(b'\x04')
    ListPairings = bytearray(b'\x05')

    # TLV Values (see table 4-6 page 61)
    kTLVType_Method = 0
    kTLVType_Identifier = 1
    kTLVType_Salt = 2
    kTLVType_PublicKey = 3
    kTLVType_Proof = 4
    kTLVType_EncryptedData = 5
    kTLVType_State = 6
    kTLVType_Error = 7
    kTLVType_RetryDelay = 8
    kTLVType_Certificate = 9
    kTLVType_Signature = 10
    kTLVType_Permissions = 11  # 0x00 => reg. user, 0x01 => admin
    kTLVType_Permission_RegularUser = bytearray(b'\x00')
    kTLVType_Permission_AdminUser = bytearray(b'\x01')
    kTLVType_FragmentData = 12
    kTLVType_FragmentLast = 13
    kTLVType_Separator = 255
    kTLVType_Separator_Pair = [255, bytearray(b'')]
    kTLVType_SessionID = 0x0e   # Table 6-27 page 116

    # Errors (see table 4-5 page 60)
    kTLVError_Unknown = bytearray(b'\x01')
    kTLVError_Authentication = bytearray(b'\x02')
    kTLVError_Backoff = bytearray(b'\x03')
    kTLVError_MaxPeers = bytearray(b'\x04')
    kTLVError_MaxTries = bytearray(b'\x05')
    kTLVError_Unavailable = bytearray(b'\x06')
    kTLVError_Busy = bytearray(b'\x07')

    # Table 6-27 page 116
    kTLVMethod_Resume = 0x07

    # Additional Parameter Types for BLE (Table 6-9 page 98)
    kTLVHAPParamValue = 0x01
    kTLVHAPParamAdditionalAuthorizationData = 0x02
    kTLVHAPParamOrigin = 0x03
    kTLVHAPParamCharacteristicType = 0x04
    kTLVHAPParamCharacteristicInstanceId = 0x05
    kTLVHAPParamServiceType = 0x06
    kTLVHAPParamServiceInstanceId = 0x07
    kTLVHAPParamTTL = 0x08
    kTLVHAPParamParamReturnResponse = 0x09
    kTLVHAPParamHAPCharacteristicPropertiesDescriptor = 0x0a
    kTLVHAPParamGATTUserDescriptionDescriptor = 0x0b
    kTLVHAPParamGATTPresentationFormatDescriptor = 0x0c
    kTLVHAPParamGATTValidRange = 0x0d
    kTLVHAPParamHAPStepValueDescriptor = 0x0e
    kTLVHAPParamHAPServiceProperties = 0x0f
    kTLVHAPParamHAPLinkedServices = 0x10
    kTLVHAPParamHAPValidValuesDescriptor = 0x11
    kTLVHAPParamHAPValidValuesRangeDescriptor = 0x12

    @staticmethod
    def decode_bytes(bs, expected=None) -> list:
        return TLV.decode_bytearray(bytearray(bs), expected)

    @staticmethod
    def decode_bytearray(ba: bytearray, expected=None) -> list:
        result = []
        # do not influence caller!
        tail = ba.copy()
        while len(tail) > 0:
            key = tail.pop(0)
            if expected and key not in expected:
                break
            length = tail.pop(0)
            value = tail[:length]
            if length != len(value):
                raise TlvParseException('Not enough data for length {} while decoding \'{}\''.format(length, ba))
            tail = tail[length:]

            if len(result) > 0 and result[-1][0] == key:
                result[-1][1] += value
            else:
                result.append([key, value])
        logger.debug('receiving %s', TLV.to_string(result))
        return result

    @staticmethod
    def validate_key(k: int) -> bool:
        try:
            val = int(k)
            if val < 0 or val > 255:
                valid = False
            else:
                valid = True
        except ValueError:
            valid = False
        return valid

    @staticmethod
    def encode_list(d: list) -> bytearray:
        logger.debug('sending %s', TLV.to_string(d))
        result = bytearray()
        for p in d:
            (key, value) = p
            if not TLV.validate_key(key):
                raise ValueError('Invalid key')

            # handle separators properly
            if key == TLV.kTLVType_Separator:
                if len(value) == 0:
                    result.append(key)
                    result.append(0)
                else:
                    raise ValueError('Separator must not have data')

            while len(value) > 0:
                result.append(key)
                if len(value) > 255:
                    length = 255
                    result.append(length)
                    for b in value[:length]:
                        result.append(b)
                    value = value[length:]
                else:
                    length = len(value)
                    result.append(length)
                    for b in value[:length]:
                        result.append(b)
                    value = value[length:]
        return result

    @staticmethod
    def to_string(d) -> str:
        def entry_to_string(entry_key, entry_value) -> str:
            if isinstance(entry_value, bytearray):
                return '  {k}: ({len} bytes/{t}) 0x{v}\n'.format(k=entry_key, v=entry_value.hex(), len=len(entry_value),
                                                                 t=type(entry_value))
            return '  {k}: ({len} bytes/{t}) {v}\n'.format(k=entry_key, v=entry_value, len=len(entry_value),
                                                           t=type(entry_value))

        if isinstance(d, dict):
            res = '{\n'
            for k in d.keys():
                res += entry_to_string(k, d[k])
            res += '}\n'
        else:
            res = '[\n'
            for k in d:
                res += entry_to_string(k[0], k[1])
            res += ']\n'
        return res

    @staticmethod
    def reorder(tlv_array, preferred_order):
        """
        This function is used to reorder the key value pairs of a TLV list according to a preferred order. If key from
        the preferred_order list is not found, it is ignored. If a pair's key is not in the preferred order list it is
        ignored as well.

        It is mostly used, if some accessory does not respect the order mentioned in the specification.

        :param tlv_array: a list of tupels containing key and value of the TLV
        :param preferred_order: a list of keys describing how the key value pairs should be sorted.
        :return: a TLV list containing only pairs whose key was in the preferred order list sorted by that order.
        """
        tmp = []
        for key in preferred_order:
            for item in tlv_array:
                if item[0] == key:
                    tmp.append(item)
        return tmp


class TlvParseException(Exception):
    """Raised upon parse error with some TLV"""
    pass



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from homekit.model.characteristics import AbstractCharacteristic, CharacteristicFormats, CharacteristicPermissions, \
    CharacteristicUnits, CharacteristicsTypes


class BrightnessCharacteristic(AbstractCharacteristic):
    """
    Defined on page 145
    """

    def __init__(self, iid):
        AbstractCharacteristic.__init__(self, iid, CharacteristicsTypes.BRIGHTNESS, CharacteristicFormats.int)
        self.perms = [CharacteristicPermissions.paired_read, CharacteristicPermissions.paired_write,
                      CharacteristicPermissions.events]
        self.description = 'Brightness in percent'
        self.minValue = 0
        self.maxValue = 100
        self.minStep = 1
        self.value = 0
        self.unit = CharacteristicUnits.percentage


class BrightnessCharacteristicMixin(object):
    def __init__(self, iid):
        self._brightnessCharacteristic = BrightnessCharacteristic(iid)
        self.characteristics.append(self._brightnessCharacteristic)

    def set_brightness_set_callback(self, callback):
        self._brightnessCharacteristic.set_set_value_callback(callback)

    def set_brightness_get_callback(self, callback):
        self._brightnessCharacteristic.set_get_value_callback(callback)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class CharacteristicFormats(object):
    """
    Values for characteristic's format taken from table 5-5 page 67
    """
    bool = 'bool'
    uint8 = 'uint8'
    uint16 = 'uint16'
    uint32 = 'uint32'
    uint64 = 'uint64'
    int = 'int'
    float = 'float'
    string = 'string'
    tlv8 = 'tlv8'
    data = 'data'


class _BleCharacteristicFormats(object):
    """
    Mapping taken from Table 6-36 page 129 and
    https://developer.nordicsemi.com/nRF5_SDK/nRF51_SDK_v4.x.x/doc/html/group___b_l_e___g_a_t_t___c_p_f___f_o_r_m_a_t_s.html
    """
    def __init__(self):
        self._formats = {
            0x01: 'bool',
            0x04: 'uint8',
            0x06: 'uint16',
            0x08: 'uint32',
            0x0A: 'uint64',
            0x10: 'int',
            0x14: 'float',
            0x19: 'string',
            0x1b: 'data'
        }

        self._formats_rev = {v: k for (k, v) in self._formats.items()}

    def get(self, key, default):
        return self._formats.get(key, default)

    def get_reverse(self, key, default):
        return self._formats_rev.get(key, default)


#
#   Have a singleton to avoid overhead
#
BleCharacteristicFormats = _BleCharacteristicFormats()



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import uuid


class _CharacteristicsTypes(object):
    """
    Translate the characteristic's UUIDs into the type description (as defined by Apple).
    E.g:
        "6D" becomes "0000006D-0000-1000-8000-0026BB765291" and translates to
        "public.hap.characteristic.position.current" or "position.current"
    Data is taken from chapter 8 of the specification (page 144ff)
    """
    ACCESSORY_PROPERTIES = 'A6'
    ACTIVE = 'B0'
    ADMINISTRATOR_ONLY_ACCESS = '1'
    AIR_PARTICULATE_DENSITY = '64'
    AIR_PARTICULATE_SIZE = '65'
    AIR_PURIFIER_STATE_CURRENT = 'A9'
    AIR_PURIFIER_STATE_TARGET = 'A8'
    AIR_QUALITY = '95'
    AUDIO_FEEDBACK = '5'
    BATTERY_LEVEL = '68'
    BRIGHTNESS = '8'
    CARBON_DIOXIDE_DETECTED = '92'
    CARBON_DIOXIDE_LEVEL = '93'
    CARBON_DIOXIDE_PEAK_LEVEL = '94'
    CARBON_MONOXIDE_DETECTED = '69'
    CARBON_MONOXIDE_LEVEL = '90'
    CARBON_MONOXIDE_PEAK_LEVEL = '91'
    CHARGING_STATE = '8F'
    COLOR_TEMPERATURE = 'CE'
    CONTACT_STATE = '6A'
    DENSITY_NO2 = 'C4'
    DENSITY_OZONE = 'C3'
    DENSITY_PM10 = 'C7'
    DENSITY_PM25 = 'C6'
    DENSITY_SO2 = 'C5'
    DENSITY_VOC = 'C8'
    DOOR_STATE_CURRENT = 'E'
    DOOR_STATE_TARGET = '32'
    FAN_STATE_CURRENT = 'AF'
    FAN_STATE_TARGET = 'BF'
    FILTER_CHANGE_INDICATION = 'AC'
    FILTER_LIFE_LEVEL = 'AB'
    FILTER_RESET_INDICATION = 'AD'
    FIRMWARE_REVISION = '52'
    HARDWARE_REVISION = '53'
    HEATING_COOLING_CURRENT = 'F'
    HEATING_COOLING_TARGET = '33'
    HORIZONTAL_TILT_CURRENT = '6C'
    HORIZONTAL_TILT_TARGET = '7B'
    HUE = '13'
    IDENTIFY = '14'
    IMAGE_MIRROR = '11F'
    IMAGE_ROTATION = '11E'
    INPUT_EVENT = '73'
    LEAK_DETECTED = '70'
    LIGHT_LEVEL_CURRENT = '6B'
    LOCK_MANAGEMENT_AUTO_SECURE_TIMEOUT = '1A'
    LOCK_MANAGEMENT_CONTROL_POINT = '19'
    LOCK_MECHANISM_CURRENT_STATE = '1D'
    LOCK_MECHANISM_LAST_KNOWN_ACTION = '1C'
    LOCK_MECHANISM_TARGET_STATE = '1E'
    LOCK_PHYSICAL_CONTROLS = 'A7'
    LOGS = '1F'
    MANUFACTURER = '20'
    MODEL = '21'
    MOTION_DETECTED = '22'
    MUTE = '11A'
    NAME = '23'
    NIGHT_VISION = '11B'
    OBSTRUCTION_DETECTED = '24'
    OCCUPANCY_DETECTED = '71'
    ON = '25'
    OUTLET_IN_USE = '26'
    PAIR_SETUP = '4C'  # new for BLE, homekit spec page 57
    PAIR_VERIFY = '4E'  # new for BLE, homekit spec page 57
    PAIRING_FEATURES = '4F'  # new for BLE, homekit spec page 58
    PAIRING_PAIRINGS = '50'  # new for BLE, homekit spec page 58
    POSITION_CURRENT = '6D'
    POSITION_HOLD = '6F'
    POSITION_STATE = '72'
    POSITION_TARGET = '7C'
    RELATIVE_HUMIDITY_CURRENT = '10'
    RELATIVE_HUMIDITY_TARGET = '34'
    ROTATION_DIRECTION = '28'
    ROTATION_SPEED = '29'
    SATURATION = '2F'
    SECURITY_SYSTEM_ALARM_TYPE = '8E'
    SECURITY_SYSTEM_STATE_CURRENT = '66'
    SECURITY_SYSTEM_STATE_TARGET = '67'
    SELECTED_RTP_STREAM_CONFIGURATION = '117'
    SERIAL_NUMBER = '30'
    SERVICE_LABEL_INDEX = 'CB'
    SERVICE_LABEL_NAMESPACE = 'CD'
    SERVICE_INSTANCE_ID = 'e604e95d-a759-4817-87d3-aa005083a0d1'.upper()  # new for BLE, homekit spec page 127
    SERVICE_SIGNATURE = 'A5'  # new for BLE, homekit spec page 128
    SETUP_ENDPOINTS = '118'
    SLAT_STATE_CURRENT = 'AA'
    SMOKE_DETECTED = '76'
    STATUS_ACTIVE = '75'
    STATUS_FAULT = '77'
    STATUS_JAMMED = '78'
    STATUS_LO_BATT = '79'
    STATUS_TAMPERED = '7A'
    STREAMING_STATUS = '120'
    SUPPORTED_AUDIO_CONFIGURATION = '115'
    SUPPORTED_RTP_CONFIGURATION = '116'
    SUPPORTED_VIDEO_STREAM_CONFIGURATION = '114'
    SWING_MODE = 'B6'
    TEMPERATURE_COOLING_THRESHOLD = 'D'
    TEMPERATURE_CURRENT = '11'
    TEMPERATURE_HEATING_THRESHOLD = '12'
    TEMPERATURE_TARGET = '35'
    TEMPERATURE_UNITS = '36'
    TILT_CURRENT = 'C1'
    TILT_TARGET = 'C2'
    TYPE_SLAT = 'C0'
    VERSION = '37'
    VERTICAL_TILT_CURRENT = '6E'
    VERTICAL_TILT_TARGET = '7D'
    VOLUME = '119'
    ZOOM_DIGITAL = '11D'
    ZOOM_OPTICAL = '11C'

    def __init__(self):
        self.baseUUID = '-0000-1000-8000-0026BB765291'
        self._characteristics = {
            '1': 'public.hap.characteristic.administrator-only-access',
            '5': 'public.hap.characteristic.audio-feedback',
            '8': 'public.hap.characteristic.brightness',
            'D': 'public.hap.characteristic.temperature.cooling-threshold',
            'E': 'public.hap.characteristic.door-state.current',
            'F': 'public.hap.characteristic.heating-cooling.current',
            '10': 'public.hap.characteristic.relative-humidity.current',
            '11': 'public.hap.characteristic.temperature.current',
            '12': 'public.hap.characteristic.temperature.heating-threshold',
            '13': 'public.hap.characteristic.hue',
            '14': 'public.hap.characteristic.identify',
            '1A': 'public.hap.characteristic.lock-management.auto-secure-timeout',
            '1C': 'public.hap.characteristic.lock-mechanism.last-known-action',
            '1D': 'public.hap.characteristic.lock-mechanism.current-state',
            '1E': 'public.hap.characteristic.lock-mechanism.target-state',
            '1F': 'public.hap.characteristic.logs',
            '19': 'public.hap.characteristic.lock-management.control-point',
            '20': 'public.hap.characteristic.manufacturer',
            '21': 'public.hap.characteristic.model',
            '22': 'public.hap.characteristic.motion-detected',
            '23': 'public.hap.characteristic.name',
            '24': 'public.hap.characteristic.obstruction-detected',
            '25': 'public.hap.characteristic.on',
            '26': 'public.hap.characteristic.outlet-in-use',
            '28': 'public.hap.characteristic.rotation.direction',
            '29': 'public.hap.characteristic.rotation.speed',
            '2F': 'public.hap.characteristic.saturation',
            '30': 'public.hap.characteristic.serial-number',
            '32': 'public.hap.characteristic.door-state.target',
            '33': 'public.hap.characteristic.heating-cooling.target',
            '34': 'public.hap.characteristic.relative-humidity.target',
            '35': 'public.hap.characteristic.temperature.target',
            '36': 'public.hap.characteristic.temperature.units',
            '37': 'public.hap.characteristic.version',
            '4C': 'public.hap.characteristic.pairing.pair-setup',  # new for BLE, homekit spec page 57
            '4E': 'public.hap.characteristic.pairing.pair-verify',  # new for BLE, homekit spec page 57
            '4F': 'public.hap.characteristic.pairing.features',  # new for BLE, homekit spec page 58
            '50': 'public.hap.characteristic.pairing.pairings',  # new for BLE, homekit spec page 58
            # new for BLE, homekit spec page 127
            'e604e95d-a759-4817-87d3-aa005083a0d1'.upper(): 'public.hap.service.protocol.service-id',
            '52': 'public.hap.characteristic.firmware.revision',
            '53': 'public.hap.characteristic.hardware.revision',
            '64': 'public.hap.characteristic.air-particulate.density',
            '65': 'public.hap.characteristic.air-particulate.size',
            '66': 'public.hap.characteristic.security-system-state.current',
            '67': 'public.hap.characteristic.security-system-state.target',
            '68': 'public.hap.characteristic.battery-level',
            '69': 'public.hap.characteristic.carbon-monoxide.detected',
            '6A': 'public.hap.characteristic.contact-state',
            '6B': 'public.hap.characteristic.light-level.current',
            '6C': 'public.hap.characteristic.horizontal-tilt.current',
            '6D': 'public.hap.characteristic.position.current',
            '6E': 'public.hap.characteristic.vertical-tilt.current',
            '6F': 'public.hap.characteristic.position.hold',
            '70': 'public.hap.characteristic.leak-detected',
            '71': 'public.hap.characteristic.occupancy-detected',
            '72': 'public.hap.characteristic.position.state',
            '73': 'public.hap.characteristic.input-event',
            '75': 'public.hap.characteristic.status-active',
            '76': 'public.hap.characteristic.smoke-detected',
            '77': 'public.hap.characteristic.status-fault',
            '78': 'public.hap.characteristic.status-jammed',
            '79': 'public.hap.characteristic.status-lo-batt',
            '7A': 'public.hap.characteristic.status-tampered',
            '7B': 'public.hap.characteristic.horizontal-tilt.target',
            '7C': 'public.hap.characteristic.position.target',
            '7D': 'public.hap.characteristic.vertical-tilt.target',
            '8E': 'public.hap.characteristic.security-system.alarm-type',
            '8F': 'public.hap.characteristic.charging-state',
            '90': 'public.hap.characteristic.carbon-monoxide.level',
            '91': 'public.hap.characteristic.carbon-monoxide.peak-level',
            '92': 'public.hap.characteristic.carbon-dioxide.detected',
            '93': 'public.hap.characteristic.carbon-dioxide.level',
            '94': 'public.hap.characteristic.carbon-dioxide.peak-level',
            '95': 'public.hap.characteristic.air-quality',
            'A5': 'public.hap.characteristic.service-signature',
            'A6': 'public.hap.characteristic.accessory-properties',
            'A7': 'public.hap.characteristic.lock-physical-controls',
            'A8': 'public.hap.characteristic.air-purifier.state.target',
            'A9': 'public.hap.characteristic.air-purifier.state.current',
            'AA': 'public.hap.characteristic.slat.state.current',
            'AB': 'public.hap.characteristic.filter.life-level',
            'AC': 'public.hap.characteristic.filter.change-indication',
            'AD': 'public.hap.characteristic.filter.reset-indication',
            'AF': 'public.hap.characteristic.fan.state.current',
            'B0': 'public.hap.characteristic.active',
            'B6': 'public.hap.characteristic.swing-mode',
            'BF': 'public.hap.characteristic.fan.state.target',
            'C0': 'public.hap.characteristic.type.slat',
            'C1': 'public.hap.characteristic.tilt.current',
            'C2': 'public.hap.characteristic.tilt.target',
            'C3': 'public.hap.characteristic.density.ozone',
            'C4': 'public.hap.characteristic.density.no2',
            'C5': 'public.hap.characteristic.density.so2',
            'C6': 'public.hap.characteristic.density.pm25',
            'C7': 'public.hap.characteristic.density.pm10',
            'C8': 'public.hap.characteristic.density.voc',
            'CB': 'public.hap.characteristic.service-label-index',
            'CD': 'public.hap.characteristic.service-label-namespace',
            'CE': 'public.hap.characteristic.color-temperature',
            '114': 'public.hap.characteristic.supported-video-stream-configuration',
            '115': 'public.hap.characteristic.supported-audio-configuration',
            '116': 'public.hap.characteristic.supported-rtp-configuration',
            '117': 'public.hap.characteristic.selected-rtp-stream-configuration',
            '118': 'public.hap.characteristic.setup-endpoints',
            '119': 'public.hap.characteristic.volume',
            '11A': 'public.hap.characteristic.mute',
            '11B': 'public.hap.characteristic.night-vision',
            '11C': 'public.hap.characteristic.zoom-optical',
            '11D': 'public.hap.characteristic.zoom-digital',
            '11E': 'public.hap.characteristic.image-rotation',
            '11F': 'public.hap.characteristic.image-mirror',
            '120': 'public.hap.characteristic.streaming-status',
        }

        self._characteristics_rev = {self._characteristics[k]: k for k in self._characteristics.keys()}

    def __getitem__(self, item):
        if item in self._characteristics:
            return self._characteristics[item]

        if item in self._characteristics_rev:
            return self._characteristics_rev[item]

        # https://docs.python.org/3.5/reference/datamodel.html#object.__getitem__ say, KeyError should be raised
        raise KeyError('Unknown Characteristic {i}?'.format(i=item))

    def get_short(self, uuid: str):
        """
        Returns the short type for a given UUID. That means that "0000006D-0000-1000-8000-0026BB765291" and "6D" both
        translates to "position.current" (after looking up "public.hap.characteristic.position.current").

        if item in self._characteristics:
            return self._characteristics[item].split('.', 3)[3]
        :param uuid: the UUID in long form or the shortened version as defined in chapter 5.6.1 page 72.
        :return: the textual representation
        """
        orig_item = uuid
        uuid = uuid.upper()
        if uuid.endswith(self.baseUUID):
            uuid = uuid.split('-', 1)[0]
            uuid = uuid.lstrip('0')

        if uuid in self._characteristics:
            return self._characteristics[uuid].split('.', maxsplit=3)[3]

        return 'Unknown Characteristic {i}'.format(i=orig_item)

    def get_short_uuid(self, item_name):
        """
        Returns the short UUID for either a full UUID or textual characteristic type name. For information on
        full and short UUID consult chapter 5.6.1 page 72 of the specification. It also supports to pass through full
        non-HomeKit UUIDs.

        :param item_name: either the type name (e.g. "public.hap.characteristic.position.current") or the short UUID as
                          string or a HomeKit specific full UUID.
        :return: the short UUID (e.g. "6D" instead of "0000006D-0000-1000-8000-0026BB765291")
        :raises KeyError: if the input is neither a UUID nor a type name. Specific error is given in the message.
        """
        orig_item = item_name
        if item_name.upper().endswith(self.baseUUID):
            item_name = item_name.upper()
            item_name = item_name.split('-', 1)[0]
            return item_name.lstrip('0')

        if item_name.upper() in self._characteristics:
            item_name = item_name.upper()
            return item_name

        if item_name.lower() in self._characteristics_rev:
            item_name = item_name.lower()
            return self._characteristics_rev[item_name]

        try:
            uuid.UUID('{{{s}}}'.format(s=item_name))
            return item_name
        except ValueError:
            raise KeyError('No short UUID found for Item {item}'.format(item=orig_item))

    def get_uuid(self, item_name):
        """
        Returns the full length UUID for either a shorted UUID or textual characteristic type name. For information on
        full and short UUID consult chapter 5.6.1 page 72 of the specification. It also supports to pass through full
        HomeKit UUIDs.

        Shorted UUID means also leading zeros are stripped.

        :param item_name: either the type name (e.g. "public.hap.characteristic.position.current") or the short UUID or
                          a HomeKit specific full UUID.
        :return: the full UUID (e.g. "0000006D-0000-1000-8000-0026BB765291")
        :raises KeyError: if the input is neither a short UUID nor a type name. Specific error is given in the message.
        """
        orig_item = item_name
        # if we get a full length uuid with the proper base and a known short one, this should also work.
        if item_name.upper().endswith(self.baseUUID):
            item_name = item_name.upper()
            item_name = item_name.split('-', 1)[0]
            item_name = item_name.lstrip('0')

        if item_name.lower() in self._characteristics_rev:
            short = self._characteristics_rev[item_name.lower()]
        elif item_name.upper() in self._characteristics:
            short = item_name.upper()
        else:
            raise KeyError('No UUID found for Item {item}'.format(item=orig_item))

        medium = '0' * (8 - len(short)) + short
        long = medium + self.baseUUID
        return long


#
#   Have a singleton to avoid overhead
#
CharacteristicsTypes = _CharacteristicsTypes()



import logging

import dbus
import gatt

from homekit.model import Categories


def _get_hci_adapter(adapter_name):
    """
    Returns the DBus interfaces of the given adapter (if the adapter exists).

    :param adapter_name: the bluetooth adapter to be returned (hci0, hci1, ...)
    :return: the existing interfaces into DBus, None if the adapter does not exist.
    """
    bus = dbus.SystemBus()
    manager = dbus.Interface(bus.get_object('org.bluez', '/'), 'org.freedesktop.DBus.ObjectManager')
    managed_objects = manager.GetManagedObjects()
    for path in managed_objects:
        ifaces = managed_objects[path]
        adapter = ifaces.get('org.bluez.Adapter1')
        if adapter is None:
            continue
        if not adapter_name or adapter_name == adapter['Address'] or path.endswith(adapter_name):
            return ifaces
    return None


def hci_adapter_exists(adapter_name):
    """
    This checks via dbus if an bluez adapter with the given name exists.

    :param adapter_name: the bluetooth adapter to be used (hci0, hci1, ...)
    :return: True if the adapter exists, False otherwise
    """
    return _get_hci_adapter(adapter_name) is not None


def hci_adapter_exists_and_supports_bluetooth_le(adapter_name):
    """
    This checks via dbus if an bluez adapter with the given name exists and if so checks if there is an interface named
    'org.bluez.LEAdvertisingManager1'. This seems to be a bit flaky but the best i've got.

    :param adapter_name: the bluetooth adapter to be used (hci0, hci1, ...)
    :return: True if the adapter exists and supports BLE, False otherwise
    """
    ifaces = _get_hci_adapter(adapter_name)
    if ifaces:
        return 'org.bluez.LEAdvertisingManager1' in ifaces
    return False


def parse_manufacturer_specific_data(input_data):
    """
    Parse the manufacturer specific data as returned via Bluez ManufacturerData. This skips the data for LEN, ADT and
    CoID as specified in Chapter 6.4.2.2 of the spec on page 124. Data therefore starts at TY (must be 0x06).

    :param input_data: manufacturer specific data as bytes
    :return: a dict containing the type (key 'type', value 'HomeKit'), the status flag (key 'sf'), human readable
             version of the status flag (key 'flags'), the device id (key 'device_id'), the accessory category
             identifier (key 'acid'), human readable version of the category (key 'category'), the global state number
             (key 'gsn'), the configuration number (key 'cn') and the compatible version (key 'cv')
    """
    logging.debug('manufacturer specific data: %s', input_data.hex())

    # the type must be 0x06 as defined on page 124 table 6-29
    ty = input_data[0]
    input_data = input_data[1:]
    if ty == 0x06:
        ty = 'HomeKit'

        ail = input_data[0]
        logging.debug('advertising interval %s', '{0:02x}'.format(ail))
        length = ail & 0b00011111
        if length != 13:
            logging.debug('error with length of manufacturer data')
        input_data = input_data[1:]

        sf = input_data[0]
        if sf == 0:
            flags = 'paired'
        elif sf == 1:
            flags = 'unpaired'
        else:
            flags = 'error'
        input_data = input_data[1:]

        device_id = (':'.join(input_data[:6].hex()[0 + i:2 + i] for i in range(0, 12, 2))).upper()
        input_data = input_data[6:]

        acid = int.from_bytes(input_data[:2], byteorder='little')
        input_data = input_data[2:]

        gsn = int.from_bytes(input_data[:2], byteorder='little')
        input_data = input_data[2:]

        cn = input_data[0]
        input_data = input_data[1:]

        cv = input_data[0]
        input_data = input_data[1:]
        if len(input_data) > 0:
            logging.debug('remaining data: %s', input_data.hex())
        return {'type': ty, 'sf': sf, 'flags': flags, 'device_id': device_id, 'acid': acid, 'gsn': gsn, 'cn': cn,
                'cv': cv, 'category': Categories[int(acid)]}

    return {'manufacturer': 'apple', 'type': ty}


# 0x004c is the Company Identifier code for Apple Inc. (see Chapter 6.4.2.2 of the spec on page 124)
COID_APPLE = 0x004c


class HomekitDiscoveryDevice(gatt.Device):
    """
    Extension to gatt.Device that uses dbus to read manufacturer specific data from bluez devices and parses this data.
    The data will be accessible via the field `homekit_discovery_data`. The device's name is stored in the field `name`.
    """

    def __init__(self, *args, **kwargs):
        gatt.Device.__init__(self, *args, **kwargs, managed=False)

        self.name = self._properties.Get('org.bluez.Device1', 'Alias')
        self.homekit_discovery_data = self._get_homekit_discovery_data()

    def _get_homekit_discovery_data(self):

        try:
            mfr_data = self._properties.Get('org.bluez.Device1', 'ManufacturerData')
        except dbus.exceptions.DBusException as e:
            if e.get_dbus_name() == 'org.freedesktop.DBus.Error.InvalidArgs':
                return {}
            raise

        # convert from dbus.Dictionary({dbus.UInt16(...): dbus.Array([dbus.Byte(...),...])}) to a dict(int: bytes)
        mfr_data = dict((int(k), bytes(bytearray(v))) for (k, v) in mfr_data.items())

        if COID_APPLE not in mfr_data:
            return {}

        parsed = parse_manufacturer_specific_data(mfr_data[COID_APPLE])

        if parsed['type'] != 'HomeKit':
            return {}

        return parsed


class HomekitDiscoveryDeviceManager(gatt.DeviceManager):
    """
    Extension to gatt.DeviceManager that checks if a new device is a HomeKit Accessory and tries to parse its HomeKit
    specific data.
    """

    def make_device(self, mac_address):
        homekit_device = HomekitDiscoveryDevice(mac_address=mac_address, manager=self)
        if not homekit_device.homekit_discovery_data:
            return
        self._manage_device(homekit_device)
        return homekit_device

    def get_devices(self):
        """
        Returns all known Bluetooth devices with explicit update.
        """
        return self._devices.values()



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'HomeKitHTTPConnection', 'HttpContentTypes', 'HttpStatusCodes'
]

from homekit.http_impl.http_client import HomeKitHTTPConnection


class _HttpContentTypes:
    JSON = 'application/hap+json'
    TLV = 'application/pairing+tlv8'


class _HttpStatusCodes:
    """
    See Table 4-2 Chapter 4.15 Page 59
    """
    OK = 200
    NO_CONTENT = 204
    MULTI_STATUS = 207
    BAD_REQUEST = 400
    FORBIDDEN = 403
    NOT_FOUND = 404
    METHOD_NOT_ALLOWED = 405
    TOO_MANY_REQUESTS = 429
    CONNECTION_AUTHORIZATION_REQUIRED = 470
    INTERNAL_SERVER_ERROR = 500

    def __init__(self):
        self._codes = {
            _HttpStatusCodes.OK: 'OK',
            _HttpStatusCodes.NO_CONTENT: 'No Content',
            _HttpStatusCodes.MULTI_STATUS: 'Multi-Status',
            _HttpStatusCodes.BAD_REQUEST: 'Bad Request',
            _HttpStatusCodes.METHOD_NOT_ALLOWED: 'Method Not Allowed',
            _HttpStatusCodes.TOO_MANY_REQUESTS: 'Too Many Requests',
            _HttpStatusCodes.CONNECTION_AUTHORIZATION_REQUIRED: 'Connection Authorization Required',
            _HttpStatusCodes.INTERNAL_SERVER_ERROR: 'Internal Server Error'
        }
        self._categories_rev = {self._codes[k]: k for k in self._codes.keys()}

    def __getitem__(self, item):
        if item in self._codes:
            return self._codes[item]

        raise KeyError('Item {item} not found'.format(item=item))


HttpStatusCodes = _HttpStatusCodes()
HttpContentTypes = _HttpContentTypes



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class CharacteristicUnits(object):
    """
    See table 5-6 page 68
    """
    celsius = 'celsius'
    percentage = 'percentage'
    arcdegrees = 'arcdegrees'
    lux = 'lux'
    seconds = 'seconds'


class _BleCharacteristicUnits(object):
    """
    Mapping taken from Table 6-37 page 130 and https://www.bluetooth.com/specifications/assigned-numbers/units
    """
    def __init__(self):
        self._formats = {
            0x272F: 'celsius',
            0x2763: 'arcdegrees',
            0x27AD: 'percentage',
            0x2700: 'unitless',
            0x2731: 'lux',
            0x2703: 'seconds',
        }

    def get(self, key, default):
        return self._formats.get(key, default)


#
#   Have a singleton to avoid overhead
#
BleCharacteristicUnits = _BleCharacteristicUnits()
